from ariths_gen.wire_components import (
    Bus
)
from ariths_gen.core.arithmetic_circuits import (
    GeneralCircuit
)
from ariths_gen.one_bit_circuits.one_bit_components import (
    HalfAdder,
    FullAdder
)
from ariths_gen.one_bit_circuits.logic_gates import (
    XorGate, NotGate
)
from ariths_gen.wire_components.wires import ConstantWireValue1


class UnsignedRippleCarrySubtractor(GeneralCircuit):
    """Class representing unsigned ripple carry subtractor.

    Unsigned ripple carry subtractor represents N-bit unsigned subtractor which is composed of
    N one bit fulladders. The first FA has carry-in set to one and the B inputs of all FAs are inverted.
    ```
          ──         ──         ──         ──        
          B3 A3      B2 A2      B1 A1      B0 A0     
          │  │       │  │       │  │       │  │      
        ┌─▼──▼─┐   ┌─▼──▼─┐   ┌─▼──▼─┐   ┌─▼──▼─┐    
        │      │ C3│      │ C2│      │ C1│      │  1
     ┌──┤  FA  │◄──┤  FA  │◄──┤  FA  │◄──┤  FA  │◄── 
     │  │      │   │      │   │      │   │      │    
     │  └──────┘   └──────┘   └──────┘   └──────┘    
     ▼     ▼          ▼          ▼          ▼        
    Cout   S3         S2         S1         S0       
    ```

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        prefix (str, optional): Prefix name of unsigned rcs. Defaults to "".
        name (str, optional): Name of unsigned rcs. Defaults to "u_rcs".
    """
    def __init__(self, a: Bus, b: Bus, prefix: str = "", name: str = "u_rcs", **kwargs):
        self.N = max(a.N, b.N)
        super().__init__(inputs=[a, b], prefix=prefix, name=name, out_N=self.N + 1, signed_out=True, **kwargs)

        # Bus sign extension in case buses have different lengths
        self.a.bus_extend(N=self.N, prefix=a.prefix)
        self.b.bus_extend(N=self.N, prefix=b.prefix)

        # Gradual addition of 1-bit adder components
        for input_index in range(self.N):
            b_not = self.add_component(NotGate(self.b.get_wire(input_index), prefix=self.prefix+"_not"+str(input_index)))
            # First adder is a half adder
            if input_index == 0:
                obj_adder = FullAdder(self.a.get_wire(input_index), b_not.out, ConstantWireValue1(), prefix=self.prefix+"_ha")
            # Rest adders are full adders
            else:
                obj_adder = FullAdder(self.a.get_wire(input_index), b_not.out, obj_adder.get_carry_wire(), prefix=self.prefix+"_fa"+str(input_index))

            self.add_component(obj_adder)
            self.out.connect(input_index, obj_adder.get_sum_wire())

        # invert the last carry wire
        self.add_component(NotGate(obj_adder.get_carry_wire(), prefix=self.prefix+"_not_c"+str(self.N-1)))
        self.out.connect(self.N, self.get_previous_component().out)


class SignedRippleCarrySubtractor(UnsignedRippleCarrySubtractor, GeneralCircuit):
    """Class representing signed ripple carry subtractor.

    Signed ripple carry subtractor represents N-bit signed subtractor which is composed of
    N one bit fulladders. The first FA has carry-in set to one and the B inputs of all FAs are inverted.
    At last XOR gates are used to ensure proper sign extension.

    Its downside is its long propagation delay the bigger the circuit is.

    ```
      __         __         __         __         __
      B3 A3      B3 A3      B2 A2      B1 A1      B0 A0
      │  │       │  │       │  │       │  │       │  │
    ┌─▼──▼─┐   ┌─▼──▼─┐   ┌─▼──▼─┐   ┌─▼──▼─┐   ┌─▼──▼─┐
    │ SIGN │ C4│      │ C3│      │ C2│      │ C1│      │  1
    │Extend│◄──┤  FA  │◄──┤  FA  │◄──┤  FA  │◄──┤  FA  │◄──
    │      │   │      │   │      │   │      │   │      │
    └──┬───┘   └──┬───┘   └──┬───┘   └──┬───┘   └──┬───┘
       ▼          ▼          ▼          ▼          ▼
      Cout        S3         S2         S1         S0
    ```

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        prefix (str, optional): Prefix name of signed rcs. Defaults to "".
        name (str, optional): Name of signed rcs. Defaults to "s_rcs".
    """
    def __init__(self, a: Bus, b: Bus, prefix: str = "", name: str = "s_rcs", **kwargs):
        self.N = max(a.N, b.N)
        super().__init__(a=a, b=b, prefix=prefix, name=name, signed=True, **kwargs)

        # Additional XOR gates to ensure correct sign extension in case of sign addition
        sign_xor_1 = XorGate(self.get_previous_component(2).a, self.get_previous_component(2).b, prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate)), parent_component=self)
        self.add_component(sign_xor_1)
        sign_xor_2 = XorGate(sign_xor_1.out, self.get_previous_component(3).get_carry_wire(), prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate)), parent_component=self)
        self.add_component(sign_xor_2)
        self.out.connect(self.N, sign_xor_2.out)



from ariths_gen.wire_components import (
    ConstantWireValue0,
    ConstantWireValue1,
    Bus
)
from ariths_gen.core.arithmetic_circuits import (
    GeneralCircuit
)
from ariths_gen.one_bit_circuits.one_bit_components import (
    FullAdder
)
from ariths_gen.multi_bit_circuits.adders import (
    UnsignedCarryLookaheadAdder
)


class CarrySaveAdderComponent(GeneralCircuit):
    """Class representing carry save adder component.

    The carry save adder component is especially useful when constructing tree multiplier architectures to reduce the propagation delay as opposed to traditional implementation of tree multipliers with half/full adders.

    The circuit is composed of full adders that operate in parallel and each take 3 input operands.
    The final output bus is composed of sum and carry bits of its corresponding contained full adders.
    ```
          C3 B3 A3   C2 B2 A2   C1 B1 A1   C0 B0 A0
         │ │ │      │ │ │      │ │ │      │ │ │
        ┌▼─▼─▼─┐   ┌▼─▼─▼─┐   ┌▼─▼─▼─┐   ┌▼─▼─▼─┐
        │      │   │      │   │      │   │      │
        │  FA  │   │  FA  │   │  FA  │   │  FA  │
        │      │   │      │   │      │   │      │
        └─┬──┬─┘   └─┬──┬─┘   └─┬──┬─┘   └─┬──┬─┘
          │c3│s3     │c2│s2     │c1│s1     │c0│s0
          └──┼──┐  ┌─┘  │┌──────┼──┼───────┘  │
             │  │  │  ┌─┼┼──────┘  └───────┐  │
             │  │  │  │ └┼──────────────┐  │  │
             └──┼──┼──┼──┼───────────┐  │  │  │
                │  │  │  │  0     0/1│  │  │  │
                │  │  │  │  │     │  │  │  │  │
                o9 o8 o7 o6 o5    o4 o3 o2 o1 o0
                c3 c2 c1 c0 0    0/1 s3 s2 s1 s0
    ```

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        c (Bus): Third input bus.
        prefix (str, optional): Prefix name of csa component. Defaults to "".
        name (str, optional): Name of csa component. Defaults to "csa_component".
        signed (bool, optional): Specifies whether the component should be used in (un)signed context (used during the construction of tree multipliers). Defaults to False.
    """
    def __init__(self, a: Bus, b: Bus, c: Bus, prefix: str = "", name: str = "csa_component", signed: bool = False, **kwargs):
        self.N = max(a.N, b.N, c.N)
        super().__init__(inputs=[a, b, c], prefix=prefix, name=name, out_N=(2*self.N)+2, signed=signed, **kwargs)
        self.out.signed = False  # CSA component has always unsigned output

        bus_extension_wire = ConstantWireValue1() if self.signed is True else ConstantWireValue0()
        self.a.bus_extend(N=self.N, prefix=a.prefix, desired_extension_wire=bus_extension_wire)
        self.b.bus_extend(N=self.N, prefix=b.prefix, desired_extension_wire=bus_extension_wire)
        self.c.bus_extend(N=self.N, prefix=c.prefix, desired_extension_wire=bus_extension_wire)

        self.sum_bits = Bus(prefix=self.prefix + "_sums", N=int(self.out.N/2))
        self.carry_bits = Bus(prefix=self.prefix + "_carries", N=int(self.out.N/2))
        self.sum_bits.connect(-1, bus_extension_wire)
        self.carry_bits.connect(0, ConstantWireValue0())

        # Gradual addition of 1-bit CSA components (FAs are used)
        for input_index in range(self.N):
            obj_adder = FullAdder(self.a.get_wire(input_index), self.b.get_wire(input_index), self.c.get_wire(input_index), prefix=self.prefix+"_fa"+str(input_index))
            self.add_component(obj_adder)

            self.sum_bits.connect(input_index, obj_adder.get_sum_wire())
            self.carry_bits.connect(input_index+1, obj_adder.get_carry_wire())

        [self.out.connect(o, self.sum_bits.get_wire(o)) for o in range(0, int(self.out.N/2))]
        [self.out.connect(o, self.carry_bits.get_wire(o-int(self.out.N/2))) for o in range(int(self.out.N/2), self.out.N)]


class UnsignedCarrySaveAdder(GeneralCircuit):
    """Class representing unsigned carry save adder.

    Unsigned carry save adder represents 3 input N-bit unsigned adder which is composed of
    N one bit full adders, that are not interconnected as in for example the ripple carry adder,
    but instead operate in parallel.

    The sum and carry bits of the individual full adders are connected to a multi bit adder to sum the final result.
    ```
       C3 B3 A3   C2 B2 A2   C1 B1 A1   C0 B0 A0
         │ │ │      │ │ │      │ │ │      │ │ │
        ┌▼─▼─▼─┐   ┌▼─▼─▼─┐   ┌▼─▼─▼─┐   ┌▼─▼─▼─┐
        │      │   │      │   │      │   │      │
        │  FA  │   │  FA  │   │  FA  │   │  FA  │
        │      │   │      │   │      │   │      │
        └─┬──┬─┘   └─┬──┬─┘   └─┬──┬─┘   └─┬──┬─┘
         ┌┘c3│s3┌────┘c2│s2┌────┘c1│s1┌────┘c0│s0
      0  │   │  │       │  │       │  │       │  0
      │Y4│X4 │Y3│X3     │Y2│X2     │Y1│X1     │Y0│X0
    ┌─▼──▼───▼──▼───────▼──▼───────▼──▼───────▼──▼─┐
    │            5-bit multi bit adder             │
    │                                              │
    └┬────────┬────────┬────────┬────────┬────────┬┘
    Cout     S4       S3       S2       S1       S0
    ```

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        c (Bus): Third input bus.
        prefix (str, optional): Prefix name of unsigned csa. Defaults to "".
        name (str, optional): Name of unsigned csa. Defaults to "u_csa".
        unsigned_adder_class_name (str, optional): Unsigned multi bit adder used to obtain final sum. Defaults to UnsignedCarryLookaheadAdder.
    """
    def __init__(self, a: Bus, b: Bus, c: Bus, prefix: str = "", name: str = "u_csa", unsigned_adder_class_name: str = UnsignedCarryLookaheadAdder, **kwargs):
        self.N = max(a.N, b.N, c.N)
        super().__init__(inputs=[a, b, c], prefix=prefix, name=name, out_N=self.N+2, **kwargs)

        # Bus sign extension in case buses have different lengths
        self.a.bus_extend(N=self.N, prefix=a.prefix)
        self.b.bus_extend(N=self.N, prefix=b.prefix)
        self.c.bus_extend(N=self.N, prefix=c.prefix)

        csa_adder = CarrySaveAdderComponent(a=self.a, b=self.b, c=self.c, prefix=prefix, name="csa_comp" + str(self.N), inner_component=True)
        self.add_component(csa_adder)

        # Obtain proper adder name with its bit width
        adder_name = unsigned_adder_class_name(a=self.a, b=self.b).prefix + str(self.N+1)
        adder_a = Bus(prefix="a", N=csa_adder.sum_bits.N)
        adder_b = Bus(prefix="b", N=csa_adder.carry_bits.N)
        adder_a.connect_bus(connecting_bus=csa_adder.out, end_connection_pos=int(csa_adder.out.N/2))
        adder_b.connect_bus(connecting_bus=csa_adder.out, start_connection_pos=int(csa_adder.out.N/2), end_connection_pos=csa_adder.out.N, offset=int(csa_adder.out.N/2))
        final_adder = unsigned_adder_class_name(a=adder_a, b=adder_b, prefix=self.prefix, name=adder_name, inner_component=True)
        self.add_component(final_adder)
        self.out.connect_bus(connecting_bus=final_adder.out)



"""
Implementation of QuAdder

For more information, see:
M. A. Hanif, R. Hafiz, O. Hasan and M. Shafique, "QuAd: Design and analysis of Quality-area optimal Low-Latency approximate Adders," 2017 54th ACM/EDAC/IEEE Design Automation Conference (DAC), Austin, TX, USA, 2017, pp. 1-6, doi: 10.1145/3061639.3062306.

"""

from ...wire_components import (
    ConstantWireValue0,
    Bus
)
from ariths_gen.core.arithmetic_circuits import (
    GeneralCircuit
)

from ariths_gen.multi_bit_circuits.adders.ripple_carry_adder import (
    UnsignedRippleCarryAdder
)
import warnings


class QuAdder(GeneralCircuit):
    """
    Implementation of QuAd

    https://ieeexplore.ieee.org/document/8060326

    The implementation is inspired by Matlab code from the authors of the paper:
    ```matlab
    temp_count=1;
    for iij=1:length(R_vect)
        fprintf(fileID,['wire [' num2str(R_vect(iij)+P_vect(iij)) ':0] temp' num2str(temp_count) ';\n']);
        temp_count=temp_count + 1;
    end

    temp_count=1;
    for iiij=1:length(R_vect)
        if (sum(R_vect(1:iiij))+P_vect(1)-1) == (sum(R_vect(1:iiij))+P_vect(1)-R_vect(iiij)-P_vect(iiij))
            fprintf(fileID,['aassign temp' num2str(temp_count) '[' num2str(R_vect(iiij)+P_vect(iiij)) ':0] = in1[' num2str(sum(R_vect(1:iiij))+P_vect(1)-1) '] + in2[' num2str(sum(R_vect(1:iiij))+P_vect(1)-1) '];\n']);
        else
            disp(R_vect(1:iiij))
            fprintf(fileID,['bassign temp' num2str(temp_count) '[' num2str(R_vect(iiij)+P_vect(iiij)) ':0] = in1[' num2str(sum(R_vect(1:iiij))+P_vect(1)-1) ':' num2str(sum(R_vect(1:iiij))+P_vect(1)-R_vect(iiij)-P_vect(iiij)) '] + in2[' num2str(sum(R_vect(1:iiij))+P_vect(1)-1) ':' num2str(sum(R_vect(1:iiij))+P_vect(1)-R_vect(iiij)-P_vect(iiij)) '];\n']);
        end
        temp_count=temp_count+1;
    end

    statement='};\n';
    temp_count=1;
    for iiij=1:length(R_vect)
        if iiij ~= length(R_vect)
            if (R_vect(iiij)==1)
                statement = [', temp' num2str(temp_count) '[' num2str(R_vect(iiij)+P_vect(iiij)-1) '] ' statement];
            else
                statement = [', temp' num2str(temp_count) '[' num2str(R_vect(iiij)+P_vect(iiij)-1) ':' num2str(P_vect(iiij)) '] ' statement];
            end
        else
            statement = ['assign res[' num2str(N) ':0] =' '{ temp' num2str(temp_count) '[' num2str(R_vect(iiij)+P_vect(iiij)) ':' num2str(P_vect(iiij)) '] ' statement];
        end
        temp_count=temp_count+1;
    end
    ```


    """

    def log(self, *args):
        if self.use_log:
            print(*args)

    def __init__(self, a, b, R, P, prefix, name="quad", adder_type=None, use_log=False, **kwargs):
        """
        :param a: Bus first input
        :param b: Bus second input
        :param R: list of integers, defines the resultant bits of all the sub-adders (the first index specifies the resultant bits of sub-adder 1 and so on)
        :param P: list of integers, defines the prediction bits of all the sub-adders (again the first index specifies the prediction bits of sub-adder 1 and so on)
        """

        if not adder_type:
            adder_type = UnsignedRippleCarryAdder

        # Assumptions checks
        assert len(R) == len(P), "R and P must have the same length"
        print([P[i] < P[i-1] + R[i-1] for i in range(1, len(P))])
        assert all([P[i] < P[i-1] + R[i-1] for i in range(1, len(P))]
                   ), "Pi must be lower than Pi-1 + Ri-1"
        assert sum(R) == a.N, "Sum of R must be equal to number of bits"

        self.use_log = use_log

        self.N = max(a.N, b.N)
        super().__init__(inputs=[a, b], prefix=prefix, name=name, out_N=self.N+1, **kwargs)

        # Bus sign extension in case buses have different lengths
        self.a.bus_extend(N=self.N, prefix=a.prefix)
        self.b.bus_extend(N=self.N, prefix=b.prefix)

        # warnings.warn("QuAdder is not tested yet")

        # Connect all outputs to zero
        for i in range(self.N+1):
            self.out[i] = ConstantWireValue0()

        # Declaration of temporary wires (just for debug purposes)
        temp_count = 0
        for iiij in range(0, len(R)):
            self.log('wire [' + str(R[iiij]+P[iiij]) +
                     ':0] temp' + str(temp_count) + ';')
            temp_count = temp_count + 1

        def bus_subconnect(out_bus, in_bus, out_indexes, in_indexes):
            out_indexes = list(out_indexes)
            in_indexes = list(in_indexes)
            assert len(out_indexes) == len(in_indexes)

            for i, j in zip(out_indexes, in_indexes):
                if j >= in_bus.N:
                    out_bus[i] = ConstantWireValue0()  # unsigned extension
                else:
                    out_bus.connect(i, in_bus.get_wire(j))  # [i] = in_bus[j]

        # Connection of adders
        temp_count = 0
        temp_bus = []
        for iiij in range(0, len(R)):
            # Former verilog output
            self.log("assign temp{}[{}:0] = in1[{}:{}] + in2[{}:{}];".format(
                temp_count,
                R[iiij]+P[iiij],
                sum(R[0:iiij + 1]) + P[0]-1,
                sum(R[0:iiij + 1]) + P[0]-R[iiij]-P[iiij],
                sum(R[0:iiij + 1]) + P[0]-1,
                sum(R[0:iiij + 1]) + P[0]-R[iiij]-P[iiij]
            ))

            a1 = Bus(f"{prefix}_temp_{temp_count}_a", R[iiij]+P[iiij])
            b1 = Bus(f"{prefix}_temp_{temp_count}_b", R[iiij]+P[iiij])

            bus_subconnect(b1, self.b,
                           range(R[iiij]+P[iiij]),
                           range(sum(R[0:iiij + 1])+P[0]-R[iiij]-P[iiij], sum(R[0:iiij + 1])+P[0]))

            bus_subconnect(a1, self.a,
                           range(R[iiij]+P[iiij]),
                           range(sum(R[0:iiij + 1])+P[0]-R[iiij]-P[iiij], sum(R[0:iiij + 1])+P[0]))

            temp_bus.append(self.add_component(
                adder_type(a1, b1, prefix=f"{prefix}_add_{temp_count}")

            ))
            temp_count = temp_count+1

        # Final connection
        temp_count = 0
        statement = "}"
        wire_id = 0
        for iiij in range(0, len(R)):
            if iiij != len(R) - 1:
                if R[iiij] == 1:
                    statement = ', temp{}[{}]'.format(
                        temp_count, R[iiij]+P[iiij] - 1) + statement
                else:
                    statement = ', temp{}[{}:{}]'.format(
                        temp_count, R[iiij]+P[iiij] - 1, P[iiij]) + statement

            else:
                statement = 'assign res[' + str(self.N) + ':0] =' + '{ temp' + str(
                    temp_count) + '[' + str(R[iiij]+P[iiij]) + ':' + str(P[iiij]) + '] ' + statement

            self.log(statement)
            for i in range(P[iiij], R[iiij]+P[iiij]):
                self.log(temp_count, i, wire_id, temp_bus[temp_count].out[i])
                self.out[wire_id] = temp_bus[temp_count].out[i]
                wire_id += 1

            temp_count = temp_count+1

        # Last carry (MSB)
        self.out[wire_id] = temp_bus[temp_count - 1].out[R[iiij]+P[iiij]]



"""
Support of custom PDK

This file defines functions for generating full and half adders
directly on the CMOS modules level.

You may add your own modules as defined in the example below
(inside `set_pdk45_library()`)

NOTE: Please call this function before calling `get_v_code_XXX()` to allow the
Verilog generation process to take into account the library's specific definitions.
"""
from ..one_bit_circuits import (
    one_bit_components
)


def set_pdk45_library():
    one_bit_components.FullAdder.use_verilog_instance = "FAX1 {unit} (.A({wirea}), .B({wireb}), .C({wirec}), .YS({wireys}), .YC({wireyc}))"
    one_bit_components.HalfAdder.use_verilog_instance = "HAX1 {unit} (.A({wirea}), .B({wireb}), .YS({wireys}), .YC({wireyc}))"
    one_bit_components.TwoOneMultiplexer.use_verilog_instance = "MUX2X1 {unit} (.A({wirea}), .B({wireb}), .S({wires}), .Y({wirey}))"



from ariths_gen.wire_components import (
    Bus
)
from ariths_gen.core.arithmetic_circuits import (
    GeneralCircuit
)
from ariths_gen.one_bit_circuits.logic_gates import (
    AndGate,
    OrGate
)
from ariths_gen.core.logic_gate_circuits import (
    TwoInputLogicGate
)


class BitReduce(GeneralCircuit):
    """Class representing tree reducer circuit. Doent work for NAND gate!
    """

    def __init__(self, a: Bus, gate: TwoInputLogicGate, prefix: str = "", name: str = "bitreduce", **kwargs):
        self.N = a.N
        super().__init__(name=name, prefix=prefix, inputs=[a], out_N=1, **kwargs)

        # tree reduction
        def create_tree(a: Bus, depth: int, branch="A"):

            #print(a)
            if a.N == 1:
                return a[0]
            else:
                half = a.N // 2
                b_in = Bus(N=half, prefix=f"b_inn{depth}A")
                c_in = Bus(N=a.N - half, prefix=f"b_inn{depth}B")
                #print(a, half, a.N)


                for i, j in enumerate(range(half)):
                    b_in[i] = a[j]
                
                for i, j in enumerate(range(half, a.N)):
                    c_in[i] = a[j]

                b = create_tree(b_in, depth=depth + 1, branch = branch + "A")
                c = create_tree(c_in, depth= depth + 1, branch = branch + "B")
                d = gate(a=b, b=c, prefix = f"{self.prefix}_red_{branch}_{depth}")
                self.add_component(d)
                return d.out
            
        sumwire = create_tree(self.a, 0, "X")
        #print(sumbus)
        self.out[0] = sumwire


class OrReduce(BitReduce):
    def __init__(self, a: Bus, prefix: str = "", name: str = "orreduce", **kwargs):
        super().__init__(a=a, gate=OrGate, prefix=prefix, name=name, **kwargs)


class AndReduce(BitReduce):
    def __init__(self, a: Bus, prefix: str = "", name: str = "andreduce", **kwargs):
        super().__init__(a=a, gate=AndGate, prefix=prefix, name=name, **kwargs)



from ariths_gen.core.logic_gate_circuits import TwoInputLogicGate, TwoInputInvertedLogicGate, OneInputLogicGate
from ariths_gen.wire_components import Wire, ConstantWireValue0, ConstantWireValue1


# Two-input #
class AndGate(TwoInputLogicGate):
    """Class representing two input AND gate.

    ```
        ┌──────┐
    ───►│  &   │
        │      ├─►
    ───►│      │
        └──────┘
    ```

    Description of the __init__ method.

    Args:
        a (Wire): First input wire.
        b (Wire): Second input wire.
        prefix (str, optional): Prefix name of AND gate. Defaults to "".
        outid (int, optional): Index of output wire. Defaults to 0.
        parent_component (object, optional) Object of upper component of which AND gate is a subcomponent. Defaults to None.
    """
    def __init__(self, a: Wire, b: Wire, prefix: str = "", outid: int = 0, parent_component: object = None):
        super().__init__(a, b, prefix, outid, parent_component)
        self.gate_type = "and_gate"
        self.cgp_function = 2
        self.operator = "&"

        # Logic gate output wire generation based on input values
        # If constant input is present, logic gate is not generated and corresponding
        # input value is propagated to the output to connect to other components
        if a.is_const() and a.value == 1:
            self.out = b
            self.disable_generation = True
        elif a.is_const() and a.value == 0:
            self.out = ConstantWireValue0()
            self.disable_generation = True
        elif b.is_const() and b.value == 1:
            self.out = a
            self.disable_generation = True
        elif b.is_const() and b.value == 0:
            self.out = ConstantWireValue0()
            self.disable_generation = True
        else:
            self.out = Wire(name=prefix)

    """ BLIF CODE GENERATION """
    def get_function_blif(self):
        """Generates Blif code representing AND gate Boolean function using its truth table.

        Returns:
            str: Blif description of AND gate's Boolean function.
        """
        if self.disable_generation:
            return f".names {self.a.get_wire_value_blif()} {self.b.get_wire_value_blif()} {self.out.name}_out\n" + \
                   f"11 1\n"
        else:
            return f".names {self.a.get_wire_value_blif()} {self.b.get_wire_value_blif()} {self.out.get_wire_value_blif()}\n" + \
                   f"11 1\n"


class NandGate(TwoInputInvertedLogicGate):
    """Class representing two input NAND gate.

    ```
        ┌──────┐
    ───►│ &    │
        │      │O──►
    ───►│      │
        └──────┘
    ```

    Description of the __init__ method.

    Args:
        a (Wire): First input wire.
        b (Wire): Second input wire.
        prefix (str, optional): Prefix name of NAND gate. Defaults to "".
        outid (int, optional): Index of output wire. Defaults to 0.
        parent_component (object, optional) Object of upper component of which NAND gate is a subcomponent. Defaults to None.
    """
    def __init__(self, a: Wire, b: Wire, prefix: str = "", outid: int = 0, parent_component: object = None):
        super().__init__(a, b, prefix, outid, parent_component)
        self.gate_type = "nand_gate"
        self.cgp_function = 5
        self.operator = "&"

        # Logic gate output wire generation based on input values
        # If constant input is present, logic gate is not generated and corresponding
        # input value is propagated to the output to connect to other components
        if a.is_const() and a.value == 1:
            assert self.parent_component, "Parent component for gate {self} is not defined"
            output = NotGate(a=b, prefix=prefix + "_not", outid=outid, parent_component=parent_component)
            self.parent_component.add_component(output) if parent_component is not None else None
            self.out = output.out
            self.disable_generation = True
        elif a.is_const() and a.value == 0:
            self.out = ConstantWireValue1()
            self.disable_generation = True
        elif b.is_const() and b.value == 1:
            assert self.parent_component, "Parent component for gate {self} is not defined"
            output = NotGate(a=a, prefix=prefix + "_not", outid=outid, parent_component=parent_component)
            self.parent_component.add_component(output) if parent_component is not None else None
            self.out = output.out
            self.disable_generation = True
        elif b.is_const() and b.value == 0:
            self.out = ConstantWireValue1()
            self.disable_generation = True
        else:
            self.out = Wire(name=prefix)

    """ BLIF CODE GENERATION """
    def get_function_blif(self):
        """Generates Blif code representing NAND gate Boolean function using its truth table.

        Returns:
            str: Blif description of NAND gate's Boolean function.
        """
        if self.disable_generation:
            return f".names {self.a.get_wire_value_blif()} {self.b.get_wire_value_blif()} {self.out.name}_out\n" + \
                   f"0- 1\n-0 1\n"
        else:
            return f".names {self.a.get_wire_value_blif()} {self.b.get_wire_value_blif()} {self.out.get_wire_value_blif()}\n" + \
                   f"0- 1\n-0 1\n"


class OrGate(TwoInputLogicGate):
    """Class representing two input OR gate.

    ```
        ┌──────┐
    ───►│ ≥1   │
        │      ├─►
    ───►│      │
        └──────┘
    ```

    Description of the __init__ method.

    Args:
        a (Wire): First input wire.
        b (Wire): Second input wire.
        prefix (str, optional): Prefix name of OR gate. Defaults to "".
        outid (int, optional): Index of output wire. Defaults to 0.
        parent_component (object, optional) Object of upper component of which OR gate is a subcomponent. Defaults to None.
    """
    def __init__(self, a: Wire, b: Wire, prefix: str = "", outid: int = 0, parent_component: object = None):
        super().__init__(a, b, prefix, outid, parent_component)
        self.gate_type = "or_gate"
        self.cgp_function = 3
        self.operator = "|"

        # Logic gate output wire generation based on input values
        # If constant input is present, logic gate is not generated and corresponding
        # input value is propagated to the output to connect to other components
        if a.is_const() and a.value == 1:
            self.out = ConstantWireValue1()
            self.disable_generation = True
        elif a.is_const() and a.value == 0:
            self.out = b
            self.disable_generation = True
        elif b.is_const() and b.value == 1:
            self.out = ConstantWireValue1()
            self.disable_generation = True
        elif b.is_const() and b.value == 0:
            self.out = a
            self.disable_generation = True
        else:
            self.out = Wire(name=prefix)

    """ BLIF CODE GENERATION """
    def get_function_blif(self):
        """Generates Blif code representing OR gate Boolean function using its truth table.

        Returns:
            str: Blif description of OR gate's Boolean function.
        """
        if self.disable_generation:
            return f".names {self.a.get_wire_value_blif()} {self.b.get_wire_value_blif()} {self.out.name}_out\n" + \
                   f"1- 1\n-1 1\n"
        else:
            return f".names {self.a.get_wire_value_blif()} {self.b.get_wire_value_blif()} {self.out.get_wire_value_blif()}\n" + \
                   f"1- 1\n-1 1\n"


class NorGate(TwoInputInvertedLogicGate):
    """Class representing two input NOR gate.

    ```
        ┌──────┐
    ───►│ ≥1   │
        │      │O──►
    ───►│      │
        └──────┘
    ```

    Description of the __init__ method.

    Args:
        a (Wire): First input wire.
        b (Wire): Second input wire.
        prefix (str, optional): Prefix name of NOR gate. Defaults to "".
        outid (int, optional): Index of output wire. Defaults to 0.
        parent_component (object, optional) Object of upper component of which NOR gate is a subcomponent. Defaults to None.
    """
    def __init__(self, a: Wire, b: Wire, prefix: str = "", outid: int = 0, parent_component: object = None):
        super().__init__(a, b, prefix, outid, parent_component)
        self.gate_type = "nor_gate"
        self.cgp_function = 6
        self.operator = "|"

        # Logic gate output wire generation based on input values
        # If constant input is present, logic gate is not generated and corresponding
        # input value is propagated to the output to connect to other components
        if a.is_const() and a.value == 1:
            self.out = ConstantWireValue0()
            self.disable_generation = True
        elif a.is_const() and a.value == 0:
            assert self.parent_component, "Parent component for gate {self} is not defined"
            output = NotGate(a=b, prefix=prefix + "_not", outid=outid, parent_component=parent_component)
            self.parent_component.add_component(output) if parent_component is not None else None
            self.out = output.out
            self.disable_generation = True
        elif b.is_const() and b.value == 1:
            self.out = ConstantWireValue0()
            self.disable_generation = True
        elif b.is_const() and b.value == 0:
            assert self.parent_component, "Parent component for gate {self} is not defined"
            output = NotGate(a=a, prefix=prefix + "_not", outid=outid, parent_component=parent_component)
            self.parent_component.add_component(output) if parent_component is not None else None
            self.out = output.out
            self.disable_generation = True
        else:
            self.out = Wire(name=prefix)

    """ BLIF CODE GENERATION """
    def get_function_blif(self):
        """Generates Blif code representing NOR gate Boolean function using its truth table.

        Returns:
            str: Blif description of NOR gate's Boolean function.
        """
        if self.disable_generation:
            return f".names {self.a.get_wire_value_blif()} {self.b.get_wire_value_blif()} {self.out.name}_out\n" + \
                   f"00 1\n"
        else:
            return f".names {self.a.get_wire_value_blif()} {self.b.get_wire_value_blif()} {self.out.get_wire_value_blif()}\n" + \
                   f"00 1\n"


class XorGate(TwoInputLogicGate):
    """Class representing two input XOR gate.

    ```
        ┌──────┐
    ───►│ =1   │
        │      ├─►
    ───►│      │
        └──────┘
    ```

    Description of the __init__ method.

    Args:
        a (Wire): First input wire.
        b (Wire): Second input wire.
        prefix (str, optional): Prefix name of XOR gate. Defaults to "".
        outid (int, optional): Index of output wire. Defaults to 0.
        parent_component (object, optional) Object of upper component of which XOR gate is a subcomponent. Defaults to None.
    """
    def __init__(self, a: Wire, b: Wire, prefix: str = "", outid: int = 0, parent_component: object = None):
        super().__init__(a, b, prefix, outid, parent_component)
        self.gate_type = "xor_gate"
        self.cgp_function = 4
        self.operator = "^"

        # Logic gate output wire generation based on input values
        # If constant input is present, logic gate is not generated and corresponding
        # input value is propagated to the output to connect to other components
        if a.is_const() and a.value == 1:
            assert self.parent_component, "Parent component for gate {self} is not defined"
            output = NotGate(a=b, prefix=prefix + "_not", outid=outid, parent_component=parent_component)
            self.parent_component.add_component(output) if parent_component is not None else None
            self.out = output.out
            self.disable_generation = True
        elif a.is_const() and a.value == 0:
            self.out = b
            self.disable_generation = True
        elif b.is_const() and b.value == 1:
            assert self.parent_component, "Parent component for gate {self} is not defined"
            output = NotGate(a=a, prefix=prefix + "_not", outid=outid, parent_component=parent_component)
            self.parent_component.add_component(output) if parent_component is not None else None
            self.out = output.out
            self.disable_generation = True
        elif b.is_const() and b.value == 0:
            self.out = a
            self.disable_generation = True
        else:
            self.out = Wire(name=prefix)

    """ BLIF CODE GENERATION """
    def get_function_blif(self):
        """Generates Blif code representing XOR gate Boolean function using its truth table.

        Returns:
            str: Blif description of XOR gate's Boolean function.
        """
        if self.disable_generation:
            return f".names {self.a.get_wire_value_blif()} {self.b.get_wire_value_blif()} {self.out.name}_out\n" + \
                   f"01 1\n10 1\n"
        else:
            return f".names {self.a.get_wire_value_blif()} {self.b.get_wire_value_blif()} {self.out.get_wire_value_blif()}\n" + \
                   f"01 1\n10 1\n"


class XnorGate(TwoInputInvertedLogicGate):
    """Class representing two input XNOR gate.

    ```
        ┌──────┐
    ───►│ =1   │
        │      │O──►
    ───►│      │
        └──────┘
    ```

    Description of the __init__ method.

    Args:
        a (Wire): First input wire.
        b (Wire): Second input wire.
        prefix (str, optional): Prefix name of XNOR gate. Defaults to "".
        outid (int, optional): Index of output wire. Defaults to 0.
        parent_component (object, optional) Object of upper component of which XNOR gate is a subcomponent. Defaults to None.
    """
    def __init__(self, a: Wire, b: Wire, prefix: str = "", outid:  int = 0, parent_component: object = None):
        super().__init__(a, b, prefix, outid, parent_component)
        self.gate_type = "xnor_gate"
        self.cgp_function = 7
        self.operator = "^"

        # Logic gate output wire generation based on input values
        # If constant input is present, logic gate is not generated and corresponding
        # input value is propagated to the output to connect to other components
        if a.is_const() and a.value == 1:
            self.out = b
            self.disable_generation = True
        elif a.is_const() and a.value == 0:
            assert self.parent_component, "Parent component for gate {self} is not defined"
            output = NotGate(a=b, prefix=prefix + "_not", outid=outid, parent_component=parent_component)
            self.parent_component.add_component(output) if parent_component is not None else None
            self.out = output.out
            self.disable_generation = True
        elif b.is_const() and b.value == 1:
            self.out = a
            self.disable_generation = True
        elif b.is_const() and b.value == 0:
            assert self.parent_component, "Parent component for gate {self} is not defined"
            output = NotGate(a=a, prefix=prefix + "_not", outid=outid, parent_component=parent_component)
            self.parent_component.add_component(output) if parent_component is not None else None
            self.out = output.out
            self.disable_generation = True
        else:
            self.out = Wire(name=prefix)

    """ BLIF CODE GENERATION """
    def get_function_blif(self):
        """Generates Blif code representing XNOR gate Boolean function using its truth table.

        Returns:
            str: Blif description of XNOR gate's Boolean function.
        """
        if self.disable_generation:
            return f".names {self.a.get_wire_value_blif()} {self.b.get_wire_value_blif()} {self.out.name}_out\n" + \
                   f"00 1\n11 1\n"
        else:
            return f".names {self.a.get_wire_value_blif()} {self.b.get_wire_value_blif()} {self.out.get_wire_value_blif()}\n" + \
                   f"00 1\n11 1\n"


# Single-input #
class NotGate(OneInputLogicGate):
    """Class representing one input NOT gate.

    ```
        ┌──────┐
        │  1   │
    ───►│      │O─►
        │      │
        └──────┘
    ```

    Description of the __init__ method.

    Args:
        a (Wire): Input wire.
        prefix (str, optional): Prefix name of NOT gate. Defaults to "".
        outid (int, optional): Index of output wire. Defaults to 0.
        parent_component (object, optional) Object of upper component of which NOT gate is a subcomponent. Defaults to None.
    """
    def __init__(self, a: Wire, prefix: str = "", outid: int = 0, parent_component: object = None):
        super().__init__(a, prefix, outid, parent_component)
        self.gate_type = "not_gate"
        self.cgp_function = 1
        self.operator = "~"

        # Logic gate output wire generation based on input values
        # If constant input is present, logic gate is not generated and corresponding
        # input value is propagated to the output to connect to other components
        if a.is_const() and a.value == 1:
            self.out = ConstantWireValue0()
            self.disable_generation = True
        elif a.is_const() and a.value == 0:
            self.out = ConstantWireValue1()
            self.disable_generation = True
        else:
            self.out = Wire(name=prefix)

    """ BLIF CODE GENERATION """
    def get_function_blif(self):
        """Generates Blif code representing NOT gate Boolean function using its truth table.

        Returns:
            str: Blif description of NOT gate's Boolean function.
        """
        if self.disable_generation:
            return f".names {self.a.get_wire_value_blif()} {self.out.name}_out\n" + \
                   f"0 1\n"
        else:
            return f".names {self.a.get_wire_value_blif()} {self.out.get_wire_value_blif()}\n" + \
                   f"0 1\n"



from ariths_gen.multi_bit_circuits.adders.carry_lookahead_adder import (
    UnsignedCarryLookaheadAdder
)
from ariths_gen.wire_components import (
    ConstantWireValue0,
    ConstantWireValue1,
    Bus
)
from ariths_gen.core.arithmetic_circuits import (
    MultiplierCircuit
)
from ariths_gen.one_bit_circuits.one_bit_components import (
    HalfAdder,
    FullAdder
)
from ariths_gen.one_bit_circuits.logic_gates import (
    XorGate
)
from ariths_gen.multi_bit_circuits.adders import (
    CarrySaveAdderComponent,
    UnsignedCarryLookaheadAdder
)


class UnsignedWallaceMultiplier(MultiplierCircuit):
    """Class representing unsigned wallace multiplier.

    Unsigned wallace multiplier represents fast N-bit multiplier which utilizes
    the functionality of wallace tree reduction algorithm proposed by Chris Wallace.

    First partial products are calculated for each bit pair that form the partial product multiplication rows/columns.
    At last the reduced pairs are inserted into the chosen multi bit unsigned adder to execute their summation and obtain the final output bits.

    The multiplier can be build from carry save adders or fully connected half/full adders (greater delay).

    The csa implementation uses carry save adder components to efficiently implement reduction of partial products utilizing the parallelism of the carry save adders. At last the reduced pairs are inserted into chosen multi bit unsigned adder to execute their summation and obtain the final output bits. It presents a faster version of multiplier opposed to the conventional architectures that are composed of interconnected half/full adders.

    Wallace tree algorithm is described more in detail here:
    https://en.wikipedia.org/wiki/Wallace_tree

    It presents a smaller circuit in area opposed to an array multiplier but is slightly bigger then dadda because of less reduction stages.
    ```
    CSA IMPLEMENTATION:

           PP7 PP6 PP5 PP4 PP3 PP2 PP1 PP0
            │   │    │  │  │     │  │  │
            │   │   ┌▼──▼──▼┐   ┌▼──▼──▼┐
            │   │   │  CSA  │   │  CSA  │
            │   │   └─┬───┬─┘   └─┬───┬─┘
            │   │     │c1 │s1     │c0 │s0
            └┐  │  ┌──┘   └────┐  │  ┌┘
            ┌▼──▼──▼┐         ┌▼──▼──▼┐
            │  CSA  │         │  CSA  │
            └─┬───┬─┘         └─┬───┬─┘
              │c4 │s4   ┌───────┘c3 │s3
              │   └──┐  │  ┌────────┘
              │     ┌▼──▼──▼┐
              │     │  CSA  │
              │     └─┬───┬─┘
              │  ┌────┘c5 │s5
              │  │  ┌─────┘
             ┌▼──▼──▼┐
             │  CSA  │
             └─┬───┬─┘
               │c6 │s6
             ┌─▼───▼─┐
             │  CPA  │
             └───┬───┘
                 o
    ```
    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        prefix (str, optional): Prefix name of unsigned wallace multiplier. Defaults to "".
        name (str, optional): Name of unsigned wallace multiplier. Defaults to "u_wallace_cla".
        use_csa (bool, optional): Choose whether to use carry save adder architecture (True) or fully interconnected half/full adders (False). Defaults to True.
        unsigned_adder_class_name (str, optional): Unsigned multi bit adder used to obtain final sums of products. Defaults to UnsignedCarryLookaheadAdder.
    """
    def __init__(self, a: Bus, b: Bus, prefix: str = "", name: str = "u_wallace_cla", use_csa: bool = True, unsigned_adder_class_name: str = UnsignedCarryLookaheadAdder, **kwargs):
        self.N = max(a.N, b.N)
        super().__init__(inputs=[a, b], prefix=prefix, name=name, out_N=self.N*2, **kwargs)

        # Bus sign extension in case buses have different lengths
        self.a.bus_extend(N=self.N, prefix=a.prefix)
        self.b.bus_extend(N=self.N, prefix=b.prefix)

        # CSA IMPLEMENTATION
        if use_csa is True:
            # Initialize all rows partial products forming AND gates matrix
            self.rows = self.init_row_lengths()

            # Zero extension of partial product rows
            for i in range(0, len(self.rows)):
                self.rows[i] = Bus(prefix=self.rows[i].prefix, wires_list=[ConstantWireValue0() for _ in range(0, i)] + self.rows[i].bus)

            while len(self.rows) > 2:
                # Gradual creation of unsigned csa adder components to reduce the pp rows to the total count of 2
                pp_index = 0
                while pp_index < len(self.rows) and (pp_index+2) < len(self.rows):
                    csa_reduction = CarrySaveAdderComponent(a=self.rows[pp_index], b=self.rows[pp_index+1], c=self.rows[pp_index+2], prefix=self.prefix+"_csa"+str(self.get_instance_num(cls=CarrySaveAdderComponent)), inner_component=True)
                    self.add_component(csa_reduction)

                    # 3 pp rows have been reduced to 2
                    [self.rows.pop(pp_index) for i in range(3)]

                    # Append rows of sum and carry results from csa calculation
                    csa_sums_N = self.out.N if csa_reduction.sum_bits.N > self.out.N-1 else csa_reduction.sum_bits.N
                    csa_sums = Bus(prefix=self.prefix+"_csa_s"+str(self.get_instance_num(cls=CarrySaveAdderComponent)), N=csa_sums_N)
                    csa_sums.connect_bus(connecting_bus=csa_reduction.out, end_connection_pos=csa_sums_N)

                    csa_carries_N = self.out.N if csa_reduction.carry_bits.N > self.out.N-1 else csa_reduction.carry_bits.N
                    csa_carries = Bus(prefix=self.prefix+"_csa_c"+str(self.get_instance_num(cls=CarrySaveAdderComponent)), N=csa_carries_N)
                    csa_carries.connect_bus(connecting_bus=csa_reduction.out, start_connection_pos=int(csa_reduction.out.N/2), end_connection_pos=int(csa_reduction.out.N/2)+csa_carries.N, offset=int(csa_reduction.out.N/2))

                    self.rows.insert(pp_index, csa_carries)
                    self.rows.insert(pp_index, csa_sums)

                    # Update of the number of pp rows
                    pp_index += 2

            # Output generation
            # 1 bit multiplier case
            if self.N == 1:
                self.out.connect(0, self.get_previous_component().out)
                self.out.connect(1, ConstantWireValue0())
            # 2 bit multiplier case
            elif self.N == 2:
                self.out.connect(0, self.components[0].out)
                obj_ha = HalfAdder(a=self.components[1].out, b=self.components[2].out, prefix=self.prefix+"_ha"+str(self.get_instance_num(cls=HalfAdder)))
                self.add_component(obj_ha)
                self.out.connect(1, obj_ha.get_sum_wire())

                obj_ha = HalfAdder(a=self.get_previous_component().get_carry_wire(), b=self.components[3].out, prefix=self.prefix+"_ha"+str(self.get_instance_num(cls=HalfAdder)))
                self.add_component(obj_ha)
                self.out.connect(2, obj_ha.get_sum_wire())
                self.out.connect(3, obj_ha.get_carry_wire())

            # Final addition of remaining bits using chosen unsigned multi bit adder
            else:
                # Obtain proper adder name with its bit width (columns bit pairs minus the first alone bit)
                adder_name = unsigned_adder_class_name(a=a, b=b).prefix + str(self.rows[0].N)
                adder_a = Bus(prefix="a", N=self.rows[0].N)
                adder_b = Bus(prefix="b", N=self.rows[1].N)
                [adder_a.connect(w, self.rows[0].get_wire(w)) for w in range(0, self.rows[0].N)]
                [adder_b.connect(w, self.rows[1].get_wire(w)) for w in range(0, self.rows[1].N)]
                final_adder = unsigned_adder_class_name(a=adder_a, b=adder_b, prefix=self.prefix, name=adder_name, inner_component=True, **kwargs)
                self.add_component(final_adder)
                [self.out.connect(o, final_adder.out.get_wire(o), inserted_wire_desired_index=o) for o in range(0, final_adder.out.N-1)]

        # FULLY INTERCONNECTED HAs/FAs IMPLEMENTATION
        else:
            # Initialize all columns partial products forming AND gates matrix
            self.columns = self.init_column_heights()

            # Perform reduction until all columns have 2 or less bits in them
            while not all(height <= 2 for (height, *_) in self.columns):
                col = 0
                while col < len(self.columns):
                    # If column has exactly 3 bits in height and all previous columns has maximum of 2 bits in height, combine them in a half adder
                    if self.get_column_height(col) == 3 and all(height <= 2 for (height, *_) in self.columns[0:col-1]):
                        # Add half adder and also AND gates if neccesarry (via add_column_wire invocation) into list of circuit components
                        obj_adder = HalfAdder(self.add_column_wire(column=col, bit=0), self.add_column_wire(column=col, bit=1), prefix=self.prefix+"_ha"+str(self.get_instance_num(cls=HalfAdder)))
                        self.add_component(obj_adder)

                        # Update the number of current and next column wires
                        self.update_column_heights(curr_column=col, curr_height_change=-1, next_column=col+1, next_height_change=1)

                        # Update current and next column wires arrangement
                        #   add ha's generated sum to the bottom of current column
                        #   add ha's generated cout to the top of next column
                        self.update_column_wires(curr_column=col, next_column=col+1, adder=self.get_previous_component(1))

                    # If column has more than 3 bits in height, combine them in a full adder
                    elif self.get_column_height(col) > 3:
                        # Add full adder and also AND gates if neccesarry (via add_column_wire invocation) into list of circuit components
                        obj_adder = FullAdder(self.add_column_wire(column=col, bit=0), self.add_column_wire(column=col, bit=1), self.add_column_wire(column=col, bit=2), prefix=self.prefix+"_fa"+str(self.get_instance_num(cls=FullAdder)))
                        self.add_component(obj_adder)

                        # Update the number of current and next column wires
                        self.update_column_heights(curr_column=col, curr_height_change=-2, next_column=col+1, next_height_change=1)

                        # Update current and next column wires arrangement
                        #   add fa's generated sum to the bottom of current column
                        #   add fa's generated cout to the top of next column
                        self.update_column_wires(curr_column=col, next_column=col+1, adder=self.get_previous_component(1))
                    col += 1

            # Output generation
            # First output bit from single first pp AND gate
            self.out.connect(0, self.add_column_wire(column=0, bit=0))
            # Final addition of remaining bits
            # 1 bit multiplier case
            if self.N == 1:
                self.out.connect(1, ConstantWireValue0())
            # 2 bit multiplier case
            elif self.N == 2:
                obj_ha = HalfAdder(self.add_column_wire(column=1, bit=0), self.add_column_wire(column=1, bit=1), prefix=self.prefix+"_ha"+str(self.get_instance_num(cls=HalfAdder)))
                self.add_component(obj_ha)
                self.out.connect(1, obj_ha.get_sum_wire())

                obj_ha = HalfAdder(self.get_previous_component().get_carry_wire(), self.add_column_wire(column=2, bit=0), prefix=self.prefix+"_ha"+str(self.get_instance_num(cls=HalfAdder)))
                self.add_component(obj_ha)
                self.out.connect(2, obj_ha.get_sum_wire())
                self.out.connect(3, obj_ha.get_carry_wire())
            # Final addition of remaining bits using chosen unsigned multi bit adder
            else:
                # Obtain proper adder name with its bit width (columns bit pairs minus the first alone bit)
                adder_name = unsigned_adder_class_name(a=a, b=b).prefix + str(len(self.columns)-1)
                adder_a = Bus(prefix=f"a", wires_list=[self.add_column_wire(column=col, bit=0) for col in range(1, len(self.columns))])
                adder_b = Bus(prefix=f"b", wires_list=[self.add_column_wire(column=col, bit=1) for col in range(1, len(self.columns))])
                final_adder = unsigned_adder_class_name(a=adder_a, b=adder_b, prefix=self.prefix, name=adder_name, inner_component=True, **kwargs)
                self.add_component(final_adder)

                [self.out.connect(o, final_adder.out.get_wire(o-1), inserted_wire_desired_index=o-1) for o in range(1, len(self.out.bus))]


class SignedWallaceMultiplier(MultiplierCircuit):
    """Class representing signed wallace multiplier.

    Signed wallace multiplier represents fast N-bit multiplier which utilizes
    the functionality of wallace tree reduction algorithm proposed by Chris Wallace and uses Baugh-Wooley algorithm
    to perform signed multiplication.

    First partial products are calculated for each bit pair that form the partial product multiplication rows/columns.
    The csa implementation uses carry save adder components to efficiently implement reduction of partial products utilizing the parallelism of the carry save adders. At last the reduced pairs are inserted into chosen multi bit unsigned adder to execute their summation and obtain the final output bits, additional XOR gate serves the necessary sign extension. It presents a faster version of multiplier opposed to the conventional architectures that are composed of interconnected half/full adders.

    Wallace tree algorithm is described more in detail here:
    https://en.wikipedia.org/wiki/Wallace_tree

    It presents a smaller circuit in area opposed to an array multiplier but is slightly bigger then dadda because of less reduction stages.

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        prefix (str, optional): Prefix name of signed wallace multiplier. Defaults to "".
        name (str, optional): Name of signed wallace multiplier. Defaults to "s_wallace_cla".
        use_csa (bool, optional): Choose whether to use carry save adder architecture (True) or fully interconnected half/full adders (False). Defaults to True.
        unsigned_adder_class_name (str, optional): Unsigned multi bit adder used to obtain final sums of products. Defaults to UnsignedCarryLookaheadAdder.
    """
    def __init__(self, a: Bus, b: Bus, prefix: str = "", name: str = "s_wallace_cla", use_csa: bool = True, unsigned_adder_class_name: str = UnsignedCarryLookaheadAdder, **kwargs):
        self.N = max(a.N, b.N)
        super().__init__(inputs=[a, b], prefix=prefix, name=name, out_N=self.N*2, signed=True, **kwargs)

        # Bus sign extension in case buses have different lengths
        self.a.bus_extend(N=self.N, prefix=a.prefix, desired_extension_wire=self.a.get_wire(self.a.N-1))
        self.b.bus_extend(N=self.N, prefix=b.prefix, desired_extension_wire=self.b.get_wire(self.b.N-1))

        # CSA IMPLEMENTATION
        if use_csa is True:
            # Initialize all rows partial products forming AND gates matrix
            self.rows = self.init_row_lengths()

            # Zero extension of partial product rows
            for i in range(0, len(self.rows)):
                self.rows[i] = Bus(prefix=self.rows[i].prefix, wires_list=[ConstantWireValue0() for _ in range(0, i)] + self.rows[i].bus)

            while len(self.rows) > 2:
                # Gradual creation of signed csa adder components to reduce the pp rows to the total count of 2
                pp_index = 0
                while pp_index < len(self.rows) and (pp_index+2) < len(self.rows):
                    csa_reduction = CarrySaveAdderComponent(a=self.rows[pp_index], b=self.rows[pp_index+1], c=self.rows[pp_index+2], prefix=self.prefix+"_csa"+str(self.get_instance_num(cls=CarrySaveAdderComponent)), inner_component=True, signed=True)
                    self.add_component(csa_reduction)

                    # 3 pp rows have been reduced to 2
                    [self.rows.pop(pp_index) for i in range(3)]

                    # Append rows of sum and carry results from csa calculation
                    csa_sums_N = self.out.N if csa_reduction.sum_bits.N > self.out.N-1 else csa_reduction.sum_bits.N
                    csa_sums = Bus(prefix=self.prefix+"_csa_s"+str(self.get_instance_num(cls=CarrySaveAdderComponent)), N=csa_sums_N)
                    csa_sums.connect_bus(connecting_bus=csa_reduction.out, end_connection_pos=csa_sums_N)

                    csa_carries_N = self.out.N if csa_reduction.carry_bits.N > self.out.N-1 else csa_reduction.carry_bits.N
                    csa_carries = Bus(prefix=self.prefix+"_csa_c"+str(self.get_instance_num(cls=CarrySaveAdderComponent)), N=csa_carries_N)
                    csa_carries.connect_bus(connecting_bus=csa_reduction.out, start_connection_pos=int(csa_reduction.out.N/2), end_connection_pos=int(csa_reduction.out.N/2)+csa_carries.N, offset=int(csa_reduction.out.N/2))

                    self.rows.insert(pp_index, csa_carries)
                    self.rows.insert(pp_index, csa_sums)

                    # Update of the number of pp rows
                    pp_index += 2

            # Output generation
            # 1 bit multiplier case
            if self.N == 1:
                self.out.connect(0, self.get_previous_component().out)
                self.out.connect(1, ConstantWireValue0())
                return
            # 2 bit multiplier case
            elif self.N == 2:
                self.out.connect(0, self.components[0].out)
                obj_ha = HalfAdder(a=self.components[1].out, b=self.components[2].out, prefix=self.prefix+"_ha"+str(self.get_instance_num(cls=HalfAdder)))
                self.add_component(obj_ha)
                self.out.connect(1, obj_ha.get_sum_wire())

                obj_fa = FullAdder(a=self.get_previous_component().get_carry_wire(), b=ConstantWireValue1(), c=self.components[3].out, prefix=self.prefix+"_fa"+str(self.get_instance_num(cls=FullAdder)))
                self.add_component(obj_fa)
                self.out.connect(2, obj_fa.get_sum_wire())
                self.out.connect(3, obj_fa.get_carry_wire())

            # Final addition of remaining bits using chosen unsigned multi bit adder
            else:
                # Obtain proper adder name with its bit width (columns bit pairs minus the first alone bit)
                adder_name = unsigned_adder_class_name(a=a, b=b).prefix + str(self.rows[0].N)
                adder_a = Bus(prefix="a", N=self.rows[0].N)
                adder_b = Bus(prefix="b", N=self.rows[1].N)
                [adder_a.connect(w, self.rows[0].get_wire(w)) for w in range(0, self.rows[0].N)]
                [adder_b.connect(w, self.rows[1].get_wire(w)) for w in range(0, self.rows[1].N)]
                final_adder = unsigned_adder_class_name(a=adder_a, b=adder_b, prefix=self.prefix, name=adder_name, inner_component=True, **kwargs)
                self.add_component(final_adder)
                [self.out.connect(o, final_adder.out.get_wire(o), inserted_wire_desired_index=o) for o in range(0, final_adder.out.N-1)]

            # Final XOR to ensure proper sign extension
            obj_xor = XorGate(ConstantWireValue1(), self.out.get_wire(self.out.N-1), prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate)), parent_component=self)
            self.add_component(obj_xor)
            self.out.connect(self.out.N-1, obj_xor.out)

        # FULLY INTERCONNECTED HAs/FAs IMPLEMENTATION
        else:
            # Initialize all columns partial products forming AND/NAND gates matrix based on Baugh-Wooley multiplication
            self.columns = self.init_column_heights()

            # Not used for 1 bit multiplier
            if self.N != 1:
                # Adding constant wire with value 1 to achieve signedness based on Baugh-Wooley multiplication algorithm
                # (adding constant value bit to last column (with one bit) to combine them in XOR gate to get the correct final multplication output bit at the end)
                self.columns[self.N].insert(1, ConstantWireValue1())
                self.update_column_heights(curr_column=self.N, curr_height_change=1)

            # Perform reduction until all columns have 2 or less bits in them
            while not all(height <= 2 for (height, *_) in self.columns):
                col = 0
                while col < len(self.columns):
                    # If column has exactly 3 bits in height and all previous columns has maximum of 2 bits in height, combine them in a half adder
                    if self.get_column_height(col) == 3 and all(height <= 2 for (height, *_) in self.columns[0:col-1]):
                        # Add half adder and also AND/NAND gates if neccesarry (via add_column_wire invocation) into list of circuit components
                        obj_adder = HalfAdder(self.add_column_wire(column=col, bit=0), self.add_column_wire(column=col, bit=1), prefix=self.prefix+"_ha"+str(self.get_instance_num(cls=HalfAdder)))
                        self.add_component(obj_adder)

                        # Update the number of current and next column wires
                        self.update_column_heights(curr_column=col, curr_height_change=-1, next_column=col+1, next_height_change=1)

                        # Update current and next column wires arrangement
                        #   add ha's generated sum to the bottom of current column
                        #   add ha's generated cout to the top of next column
                        self.update_column_wires(curr_column=col, next_column=col+1, adder=self.get_previous_component(1))

                    # If column has more than 3 bits in height, combine them in a full adder
                    elif self.get_column_height(col) > 3:
                        # Add full adder and also AND/NAND gates if neccesarry (via add_column_wire invocation) into list of circuit components
                        obj_adder = FullAdder(self.add_column_wire(column=col, bit=0), self.add_column_wire(column=col, bit=1), self.add_column_wire(column=col, bit=2), prefix=self.prefix+"_fa"+str(self.get_instance_num(cls=FullAdder)))
                        self.add_component(obj_adder)

                        # Update the number of current and next column wires
                        self.update_column_heights(curr_column=col, curr_height_change=-2, next_column=col+1, next_height_change=1)

                        # Update current and next column wires arrangement
                        #   add fa's generated sum to the bottom of current column
                        #   add fa's generated cout to the top of next column
                        self.update_column_wires(curr_column=col, next_column=col+1, adder=self.get_previous_component(1))
                    col += 1

            # Output generation
            # First output bit from single first pp AND gate
            self.out.connect(0, self.add_column_wire(column=0, bit=0))
            # Final addition of remaining bits
            # 1 bit multiplier case
            if self.N == 1:
                self.out.connect(1, ConstantWireValue0())
                return
            # 2 bit multiplier case
            elif self.N == 2:
                obj_ha = HalfAdder(self.add_column_wire(column=1, bit=0), self.add_column_wire(column=1, bit=1), prefix=self.prefix+"_ha"+str(self.get_instance_num(cls=HalfAdder)))
                self.add_component(obj_ha)
                self.out.connect(1, obj_ha.get_sum_wire())

                obj_fa = FullAdder(self.get_previous_component().get_carry_wire(), self.add_column_wire(column=2, bit=0), self.add_column_wire(column=2, bit=1), prefix=self.prefix+"_fa"+str(self.get_instance_num(cls=FullAdder)))
                self.add_component(obj_fa)
                self.out.connect(2, obj_fa.get_sum_wire())
                self.out.connect(3, obj_fa.get_carry_wire())

            # Final addition of remaining bits using chosen unsigned multi bit adder
            else:
                # Obtain proper adder name with its bit width (columns bit pairs minus the first alone bit)
                adder_name = unsigned_adder_class_name(a=a, b=b).prefix + str(len(self.columns)-1)
                adder_a = Bus(prefix=f"a", wires_list=[self.add_column_wire(column=col, bit=0) for col in range(1, len(self.columns))])
                adder_b = Bus(prefix=f"b", wires_list=[self.add_column_wire(column=col, bit=1) for col in range(1, len(self.columns))])
                final_adder = unsigned_adder_class_name(a=adder_a, b=adder_b, prefix=self.prefix, name=adder_name, inner_component=True, **kwargs)
                self.add_component(final_adder)

                [self.out.connect(o, final_adder.out.get_wire(o-1), inserted_wire_desired_index=o-1) for o in range(1, len(self.out.bus))]

            # Final XOR to ensure proper sign extension
            obj_xor = XorGate(ConstantWireValue1(), self.out.get_wire(self.out.N-1), prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate)), parent_component=self)
            self.add_component(obj_xor)
            self.out.connect(self.out.N-1, obj_xor.out)



from .two_input_one_bit_circuit import (
    TwoInputOneBitCircuit
)
from ariths_gen.core.arithmetic_circuits import (
    GeneralCircuit
)
from ariths_gen.wire_components.wires import Wire


class FourInputOneBitCircuit(TwoInputOneBitCircuit, GeneralCircuit):
    """Class represents a general four input one bit circuit and implements their generation to various representations. It is derived from `TwoInputOneBitCircuit` class.

    Description of the __init__ method.

    Args:
        a (Wire): First input wire.
        b (Wire): Second input wire.
        c (Wire): Third input wire.
        d (Wire): Fourth input wire.
        prefix (str, optional): Prefix name of circuit. Defaults to "".
        name (str, optional): Name of circuit. Defaults to "four_input_one_bit_circuit".
    """
    def __init__(self, a: Wire = Wire(name="a"), b: Wire = Wire(name="b"), c: Wire = Wire(name="c"), d: Wire = Wire(name="d"), prefix: str = "", name: str = "four_input_one_bit_circuit"):
        GeneralCircuit.__init__(self, inputs=[a, b, c, d], prefix=prefix, name=name, out_N=1)
        self.c_data_type = "uint8_t"

    """ C CODE GENERATION """
    # FLAT C #
    # Function prototype with four inputs
    def get_prototype_c(self):
        """Generates C code function header to describe corresponding four input one bit circuit's interface in C code.

        Returns:
            str: Function's name and parameters in C code.
        """
        return f"{self.c_data_type} {self.prefix}({self.c_data_type} {self.a.prefix}, {self.c_data_type} {self.b.prefix}, {self.c_data_type} {self.c.prefix}, {self.c_data_type} {self.d.prefix})" + "{" + "\n"

    # HIERARCHICAL C #
    # Subcomponent generation (four inputs)
    def get_out_invocation_c(self):
        """Generates hierarchical C code invocation of corresponding four input one bit circuit's generated function block.

        Assigns output values from invocation of the corresponding function block into inner wires present inside the upper
        component from which function block has been invoked.

        Returns:
            str: Hierarchical C code subcomponent's C function invocation and output assignment.
        """
        # Used to retrieve proper component's output wire offset position within the output bus
        output_bus_wire_names = []
        [output_bus_wire_names.append(w.prefix) for w in self.out.bus]
        circuit_block = self.__class__()
        return "".join([f"  {c.out.prefix} = ({circuit_block.prefix}({self.a.get_wire_value_c_hier()}, {self.b.get_wire_value_c_hier()}, {self.c.get_wire_value_c_hier()}, {self.d.get_wire_value_c_hier()}) >> {output_bus_wire_names.index(c.out.prefix)}) & 0x01;\n" for c in self.components if c.disable_generation is False and c.out.prefix in output_bus_wire_names])

    """ VERILOG CODE GENERATION """
    # FLAT VERILOG #
    # Module prototype with four inputs
    def get_prototype_v(self):
        """Generates Verilog for flat module header to describe corresponding four input one bit circuit's interface in Verilog.

        It is adapted for generation of general description of four input one bit circuits as well as their modified versions when some inputs are desired as constant values.
        In such cases the inner module logic is also modified accordingly. It is used for self four input one bit circuit flat generation.

        Returns:
            str: Flat module's name and parameters in Verilog.
        """
        unique_out_wires = []
        [unique_out_wires.append(o.name+"_outid"+str(self.out.bus.index(o))) if o.is_const() or o.name in [self.a.name, self.b.name, self.c.name, self.d.name] else unique_out_wires.append(o.name) for o in self.out.bus]
        return f"module {self.prefix}(input {self.a.name}, input {self.b.name}, input {self.c.name}, input {self.d.name}" + \
               "".join([f", output {o}" for o in unique_out_wires]) + ");\n"

    def get_self_init_v_flat(self):
        """Generates Verilog for self flat module initialization and assignment of corresponding four input one bit circuit's input/output wires.

        It is adapted for generation of general description of four input one bit circuits as well as their modified versions when some inputs are desired as constant values.
        In such cases the inner module prototype is also modified accordingly. It is used for self four input one bit circuit flat generation.

        Returns:
            str: Verilog flat module's inner circuit wires initialization and assignment.
        """
        unique_out_wires = []
        [unique_out_wires.append(o.name+"_outid"+str(self.out.bus.index(o))) if o.is_const() or o.name in [self.a.name, self.b.name, self.c.name, self.d.name] else unique_out_wires.append(o.name) for o in self.out.bus]
        return "".join([c.get_assign_v_flat() if c.disable_generation is False else
                        f"  assign {unique_out_wires.pop(unique_out_wires.index(c.out.name+'_outid'+str(c.outid)))} = {c.out.v_const};\n" if f"{c.out.name+'_outid'+str(c.outid)}" in unique_out_wires and c.out.is_const() else
                        f"  assign {unique_out_wires.pop(unique_out_wires.index(c.out.name+'_outid'+str(c.outid)))} = {c.out.name};\n" if f"{c.out.name+'_outid'+str(c.outid)}" in unique_out_wires else
                        f"" for c in self.components])

    # HIERARCHICAL VERILOG #
    # Self circuit hierarchical generation
    def get_prototype_v_hier(self):
        """Generates Verilog for hierarchical module header to describe corresponding four input one bit circuit's interface in Verilog.

        It is adapted for generation of general description of four input one bit circuits as well as their modified versions when some inputs are desired as constant values.
        In such cases the inner module logic is also modified accordingly. It is used for self four input one bit circuit hierarchical generation.

        Returns:
            str: Hierarchical module's name and parameters in Verilog.
        """
        unique_out_wires = []
        [unique_out_wires.append(o.name+"_outid"+str(self.out.bus.index(o))) if o.is_const() or o.name in [self.a.name, self.b.name, self.c.name, self.d.name] else unique_out_wires.append(o.name) for o in self.out.bus]
        return f"module {self.prefix}(input [0:0] {self.a.name}, input [0:0] {self.b.name}, input [0:0] {self.c.name}, input [0:0] {self.d.name}" + \
               "".join([f", output [0:0] {o}" for o in unique_out_wires]) + ");\n"

    def get_self_init_v_hier(self):
        """Generates Verilog for hierarchical module's inner initialization and assignment of corresponding arithmetic circuit's input/output wires.

        It is adapted for generation of general description of four input one bit circuits as well as their modified versions when some inputs are desired as constant values.
        In such cases the inner module prototype is also modified accordingly. It is used for self four input one bit circuit hierarchical generation.

        Returns:
            str: Verilog hierarchical module's inner circuit wires initialization and assignment.
        """
        unique_out_wires = []
        [unique_out_wires.append(o.name+"_outid"+str(self.out.bus.index(o))) if o.is_const() or o.name in [self.a.name, self.b.name, self.c.name, self.d.name] else unique_out_wires.append(o.name) for o in self.out.bus]
        return "".join([c.get_gate_invocation_v() if c.disable_generation is False else
                        f"  assign {unique_out_wires.pop(unique_out_wires.index(c.out.name+'_outid'+str(c.outid)))}[0] = {c.out.v_const};\n" if f"{c.out.name+'_outid'+str(c.outid)}" in unique_out_wires and c.out.is_const() else
                        f"  assign {unique_out_wires.pop(unique_out_wires.index(c.out.name+'_outid'+str(c.outid)))}[0] = {c.out.name}[0];\n" if f"{c.out.name+'_outid'+str(c.outid)}" in unique_out_wires else
                        f"" for c in self.components])

    # Subcomponent generation
    def get_out_invocation_v(self):
        """Generates hierarchical Verilog code invocation of corresponding four input one bit circuit's generated function block.

        Assigns output values from invocation of the corresponding function block into inner wires present inside the upper
        component from which function block has been invoked.

        Returns:
            str: Hierarchical Verilog code subcomponent's module invocation and output assignment.
        """
        circuit_block = self.__class__()
        return f"  {circuit_block.prefix} {circuit_block.prefix}_{self.out.prefix}(.{circuit_block.a.prefix}({self.a.get_wire_value_v_hier()}), .{circuit_block.b.prefix}({self.b.get_wire_value_v_hier()}), .{circuit_block.c.prefix}({self.c.get_wire_value_v_hier()}), .{circuit_block.d.prefix}({self.d.get_wire_value_v_hier()}){self.out.get_unique_assign_out_wires_v(circuit_block)});\n"

    """ BLIF CODE GENERATION """
    # FLAT BLIF #
    # Model prototype with four inputs
    def get_declaration_blif(self):
        """Generates Blif code declaration of four input one bit circuit's input/output wires.

        It is adapted for generation of general description of four input one bit circuits as well as their modified versions when some inputs are desired as constant values.
        In such cases the inner modul logic is also modified accordingly. It is used for self four input one bit circuit flat/hierarchical generation.

        Returns:
            str: Blif code containing declaration of circuit's input/output wires.
        """
        unique_out_wires = []
        [unique_out_wires.append(o.name+"_outid"+str(self.out.bus.index(o))) if o.is_const() or o.name in [self.a.name, self.b.name, self.c.name, self.d.name] else unique_out_wires.append(o.name) for o in self.out.bus]
        return f".inputs {self.a.get_wire_declaration_blif()}{self.b.get_wire_declaration_blif()}{self.c.get_wire_declaration_blif()}{self.d.get_wire_declaration_blif()}\n" + \
               f".outputs" + \
               "".join([f" {o}" for o in unique_out_wires]) + "\n" + \
               f".names vdd\n1\n" + \
               f".names gnd\n0\n"

    def get_function_blif_flat(self, top_modul: bool = False):
        """Generates flat Blif code with invocation of subcomponents logic gates Boolean functions via their corresponding truth tables.

        It is adapted for generation of general description of four input one bit circuits as well as their modified versions when some inputs are desired as constant values.
        In such cases the inner modul prototype is also modified accordingly. It is used for self four input one bit circuit flat generation.

        Args:
            top_modul (bool, optional): Specifies whether the described circuit represents top modul component (self one bit circuit generation). Defaults to False.

        Returns:
            str: Flat Blif code containing invocation of inner subcomponents logic gates Boolean functions.
        """
        if top_modul:
            unique_out_wires = []
            [unique_out_wires.append(o.name+"_outid"+str(self.out.bus.index(o))) if o.is_const() or o.name in [self.a.name, self.b.name, self.c.name, self.d.name] else unique_out_wires.append(o.name) for o in self.out.bus]
            return "".join([c.get_function_blif_flat() if c.disable_generation is False else
                           c.out.get_assign_blif(prefix=f"{unique_out_wires.pop(unique_out_wires.index(c.out.name+'_outid'+str(c.outid)))}", output=True) if f"{c.out.name+'_outid'+str(c.outid)}" in unique_out_wires else
                           "" for c in self.components])
        else:
            return "".join([c.get_function_blif_flat() for c in self.components])

    # HIERARCHICAL BLIF #
    # Subcomponent generation (4 inputs)
    def get_invocations_blif_hier(self):
        """Generates hierarchical Blif code with invocation of subcomponents function blocks.

        It is adapted for generation of general description of four input one bit circuits as well as their modified versions when some inputs are desired as constant values.
        In such cases the inner modul prototype is also modified accordingly. It is used for self four input one bit circuit hierarchical generation.

        Returns:
            str: Hierarchical Blif code containing invocation of inner subcomponents function blocks.
        """
        unique_out_wires = []
        [unique_out_wires.append(o.name+"_outid"+str(self.out.bus.index(o))) if o.is_const() or o.name in [self.a.name, self.b.name, self.c.name, self.d.name] else unique_out_wires.append(o.name) for o in self.out.bus]
        return "".join([c.get_invocation_blif_hier() if c.disable_generation is False else
                       c.out.get_assign_blif(prefix=f"{unique_out_wires.pop(unique_out_wires.index(c.out.name+'_outid'+str(c.outid)))}", output=True) if f"{c.out.name+'_outid'+str(c.outid)}" in unique_out_wires else
                       "" for c in self.components])

    def get_invocation_blif_hier(self):
        """Generates hierarchical Blif code invocation of corresponding four input one bit circuit's generated function block.

        Returns:
            str: Hierarchical Blif code subcomponent's model invocation.
        """
        circuit_block = self.__class__()
        return f".subckt {circuit_block.prefix} {circuit_block.a.prefix}={self.a.get_wire_value_blif()} {circuit_block.b.prefix}={self.b.get_wire_value_blif()} {circuit_block.c.prefix}={self.c.get_wire_value_blif()} {circuit_block.d.prefix}={self.d.get_wire_value_blif()}{self.out.get_unique_assign_out_wires_blif(function_block_out_bus=circuit_block.out)}\n"

    """ CGP CODE GENERATION """
    # FLAT CGP #
    # Chromosome prototype with four inputs
    def get_parameters_cgp(self):
        """Generates CGP chromosome parameters of corresponding four input one bit circuit.

        In total seven parameters represent: total inputs, total outputs, number of rows, number of columns (gates),
        number of each gate's inputs, number of each gate's outputs, quality constant value.

        Returns:
            str: CGP chromosome parameters of described circuit.
        """
        self.circuit_gates = self.get_circuit_gates()
        return f"{{4,{self.out.N},1,{len(self.circuit_gates)},2,1,0}}"



from ariths_gen.wire_components import (
    Bus
)
from ariths_gen.core.arithmetic_circuits import (
    GeneralCircuit
)

from ariths_gen.multi_bit_circuits.others import (
    UnsignedPopCount,
    UnsignedCompareGTE
)


class PopCountCompare(GeneralCircuit):
    """Class representing a circiut
      if number of ones in a is larger or equal than number of ones in b

         │ │ │ │         │ │ │ │ │ 
         ▼ ▼ ▼ ▼         ▼ ▼ ▼ ▼ ▼ 
        ┌───────┐       ┌─────────┐
        │ popcnt│       │ popcnt  │
        └──┬────┘       └─────┬───┘
           │                  │    
           └───────┐   ┌──────┘    
                ┌──▼───▼──┐         
                │   <=    │         
                └────┬────┘         
                     │             
                     ▼             
    """

    def __init__(self, a: Bus, b: Bus, prefix : str = "", name : str = "popcnt_cmp", **kwargs):
        super().__init__(name=name, prefix=prefix, inputs=[a, b], out_N=1, **kwargs)

        p1 = self.add_component(UnsignedPopCount(a=Bus(wires_list=self.a.bus, prefix=f"{prefix}_popcount1_a"),
                                                 prefix=f"{prefix}_popcount1",
                                                 inner_component=True)).out
        p2 = self.add_component(UnsignedPopCount(a=Bus(wires_list=self.b.bus, prefix=f"{prefix}_popcount2_a"),
                                                 prefix=f"{prefix}_popcount2",
                                                 inner_component=True)).out
        #N = max(p1.N, p2.N)
        #p1.bus_extend(N)
        #p2.bus_extend(N)
        red = self.add_component(UnsignedCompareGTE(p1, p2, prefix=f"{prefix}_cmp", inner_component = True))
        self.out.connect_bus(red.out)



from ariths_gen.core.logic_gate_circuits.logic_gate_circuit import (
    OneInputLogicGate,
    TwoInputLogicGate
)
from ariths_gen.wire_components import (
    Wire,
    Bus
)
from typing import Dict
import inspect
import copy
from io import StringIO


class GeneralCircuit():
    """Class represents a general circuit and ensures its generation to various representations.

    The __init__ method fills some mandatory attributes concerning arithmetic circuit
    that are later used for generation into various representations.
    """
    def __init__(self, prefix: str, name: str, out_N: int, inner_component: bool = False, inputs: list = [], one_bit_circuit: bool = False, signed: bool = False, signed_out = None, outname=None, **kwargs):
        if prefix == "":
            self.prefix = name
        else:
            self.prefix = prefix + "_" + name
        self.inner_component = inner_component

        # Dynamic input bus assignment
        self.inputs = []
        for i, input in enumerate(inputs):
            attr_name = chr(97+i)
            full_prefix = f"{self.prefix}_{input.prefix}" if self.inner_component else f"{input.prefix}"
            if isinstance(input, Bus) or isinstance(input, Wire):
                circuit_input = input
                circuit_input.prefix = full_prefix
            setattr(self, attr_name, circuit_input)
            self.inputs.append(circuit_input)
            # If the input bus is an output bus, connect it
            if isinstance(input, Bus) and input.is_output_bus():
                getattr(self, attr_name).connect_bus(connecting_bus=input)

        if not outname:
            outname = self.prefix+"_out"

        if signed_out is None:
            signed_out = signed
        self.out = Bus(outname, out_N, out_bus=True, signed=signed_out)

        self.components = []
        self._prefixes = [] # TODO rename to fullname and add distinct attr for prefix, name, suffix
        self.circuit_gates = []
        self.circuit_wires = []
        self.signed = signed
        self.signed_out = signed_out
        self.c_data_type = "int64_t" if self.signed is True else "uint64_t"
        self.pyc = None  # Python compiled function
        self.kwargs = kwargs

    def __call__(self, *args):
        if not self.pyc:
            buf = StringIO()
            self.get_python_code_flat(buf)

            globs, locs = {}, {}
            exec(buf.getvalue(), globs, locs)
            self.pyc = locs[self.prefix]

        return self.pyc(*args)

    def __str__(self):
        return f"<{type(self).__name__} prefix={self.prefix} " + (", ".join([f"input={i}" for i in self.inputs])) + ">"
        # super().__init__(prefix, name, out_N, inner_component, inputs=[a, b], signed=signed, **kwargs)

    def get_hier_subcomponent_def(self, parent_kwargs: dict = {}):
        """ Creates and returns a new instance of the current circuit block used for definition of a subcomponent in a hierarchical circuit.

        Args:
            parent_kwargs (dict): Dictionary containing all the configuration settings of the parent circuit block.

        Returns:
            GeneralCircuit: A new instance of the current circuit block with proper prefix and input wires.
        """
        # Obtain proper circuit name with its input bit widths
        init_signature = inspect.signature(self.__class__.__init__)
        init_params = list(init_signature.parameters.keys())
        default_circuit_name = init_signature.parameters['name'].default
        circuit_type = default_circuit_name + "x".join(str(getattr(self, chr(97+i)).N) for i, _ in enumerate(self.inputs))
        # Initialize and fill args for the new instance based on the current instance
        init_args = {}

        for param in init_params[1:]:  # Skip 'self'
            attr = getattr(self, param, None)  # Get the attribute from the current instance

            if attr is not None:  # If attribute does not exist, it will use default value from the signature
                if isinstance(attr, Bus):  # If the input is a Bus, create a copy of the Bus object with same length, but proper prefix
                    init_args[param] = Bus(N=attr.N, prefix=param)
                elif isinstance(attr, Wire):  # If the input is a Wire, create a copy of the Wire object with proper prefix
                    init_args[param] = Wire(name=param)
                else:  # Copy other types of attributes
                    init_args[param] = copy.deepcopy(attr)

        init_args['name'] = circuit_type
        init_args['prefix'] = ""

        circuit_block = self.__class__(**init_args, **parent_kwargs)
        return circuit_block

    def get_circuit_def(self) -> Dict[str, Wire]:
        """ returns IDs and wires of the inputs and output"""
        # TODO delete? (probably replaced by get_hier_subcomponent_def)
        #.{circuit_block.a.prefix}({self.a.prefix}), .{circuit_block.b.prefix}({self.b.prefix}), .{circuit_block.out.prefix}({self.out.prefix}));\n"
        r = {chr(97 + i): self.inputs[i] for i in range(len(self.inputs))}
        r['out'] = self.get_global_prefix() + "_out"
        return r

    def add_component(self, component):
        """Adds a component into list of circuit's inner subcomponents.
        
        Additionally it adds all the gates of the component to the circuit's list of gates and all
        sbcomponents prefixes to check for naming duplicates which could cause issues in the circuit generation.

        Args:
            component: Subcomponent to be added into list of components composing described circuit.
        """
        # TODO should be refactored in ArithsGen rework
        # We should probably check also wire names for especially hierarchical generation
        if isinstance(component, TwoInputLogicGate):
            if component.disable_generation is False:
                self.circuit_gates.append(component)
        else:
            self.circuit_gates.extend(component.get_circuit_gates())
            for prefix in component._prefixes:
                assert prefix not in self._prefixes, f"Component with prefix {prefix} already exists in the circuit."
            self._prefixes.extend(component._prefixes)

        assert component.prefix not in self._prefixes, f"Component with prefix {component.prefix} already exists in the circuit."
        self._prefixes.append(component.prefix)
        self.components.append(component)
        return component

    def get_previous_component(self, number: int = 1):
        """Retrieves previously added composite subcomponent from circuit's list of components.

        Args:
            number (int, optional): Offset indicating which lastly added component will be retrieved. Defaults to 1.

        Returns:
            component: Desired previously added composite component.
        """
        return self.components[-number]

    def get_instance_num(self, cls, count_disabled_gates: bool = True):
        """Informs how many instances of the same type are already present inside circuit's components list.

        Args:
            cls (type): Class type for which to count the number of instances in the components list.
            count_disabled_gates (bool, optional): Indicates whether logic gates that aren't generated should be also counted. Defaults to True.
        Returns:
            int: Number of instances of the same class type.
        """
        if issubclass(cls, TwoInputLogicGate) and count_disabled_gates is False:
            return sum(isinstance(c, cls) for c in self.components if isinstance(c, cls) and c.disable_generation is False)
        else:
            return sum(isinstance(c, cls) for c in self.components)
        
    def get_circuit_gates(self, verilog_output: bool = False):
        """Gets a list of all the logic gates in circuit that should be generated.

        Args:            
            verilog_output (bool): Specifies whether the call has been invoked by a verilog output generation method.
        Returns:
            list: List of composite logic gates.
        """
        gates = []
        for c in self.components:
            if isinstance(c, TwoInputLogicGate):
                if (c.disable_generation is False) and (verilog_output is False or getattr(c, "use_verilog_instance", False) is False):
                    gates.append(c)
            else:
                # Check whether it is necessary to use gates for the Verilog component
                # description (ArithsGen internally defined comp) or not (technology specific instance)
                if verilog_output is False or ((hasattr(c, "use_verilog_instance") and c.use_verilog_instance is False) or hasattr(c, "use_verilog_instance") is False):
                    gates.extend(c.get_circuit_gates(verilog_output))
        return gates

    def get_one_bit_components(self):
        """Retrieves a list of all the one bit circuits (besides logic gates) present as subcomponents inside the circuit.

        Returns:
            list: List of composite one bit circuits.
        """
        one_bit_comps = []
        for c in self.components:
            if isinstance(c, TwoInputLogicGate):
                continue
            elif all(isinstance(i, Wire) for i in self.inputs):
                one_bit_comps.append(c)
            else:
                one_bit_comps.extend(c.get_one_bit_components())

        return one_bit_comps

    def get_multi_bit_components(self):
        """Retrieves a list of all the multi bit circuits present as subcomponents inside the circuit.

        Returns:
            list: List of composite multi bit circuits.
        """
        multi_bit_comps = []
        for c in self.components:
            if isinstance(c, TwoInputLogicGate):
                continue
            elif all(isinstance(i, Wire) for i in self.inputs):
                continue
            else:
                multi_bit_comps.append(c)
                multi_bit_comps.extend(c.get_multi_bit_components())
        return multi_bit_comps

    @staticmethod
    def get_unique_types(components: list, multi_bit: bool = False):
        """Retrieves just the unique representatives of class types present inside the provided components list.

        Args:
            components (list): List of components to be filtered.
            multi_bit (bool): Specifies whether the provided components list is composed of multi bit type circuits. Defaults to False.

        Returns:
            list: List of unique composite class types.
        """
        if multi_bit is True:
            return list({(type(c), tuple(i.N for i in c.inputs)): c for c in components[::-1]}.values())
        else:
            return list({type(c): c for c in components}.values())

    def get_component_types(self, verilog_output: bool = False):
        """Retrieves a list of all the unique types of subcomponents composing the circuit.

        Returning list consists of only the unique types of logic gates, one bit circuits and multi bit circuits.

        Args:
            verilog_output (bool): Specifies whether the call has been invoked by a verilog output generation method.
        Returns:
            list: List of unique component types describing the circuit.
        """
        gate_comps = self.get_unique_types(components=self.get_circuit_gates(verilog_output))
        one_bit_comps = self.get_unique_types(
            components=self.get_one_bit_components())
        multi_bit_comps = self.get_unique_types(components=self.get_multi_bit_components(),
                                                multi_bit=True)

        all_components = gate_comps + one_bit_comps + multi_bit_comps
        return all_components

    def save_wire_id(self, wire: Wire):
        """Returns appropriate wire index position within the circuit.

        Constant wire with value 0 has constant index of 0.
        Constant wire with value 1 has constant index of 1.
        Other wires indexes start counting from 2 and up.

        Args:
            wire (Wire): Wire that will be stored at this circuit index position.

        Returns:
            int: Wire's index position within circuit.
        """
        if wire.is_const():
            return wire.cgp_const
        else:
            # [1] is reservation for a constant wire with value 1
            pos = max([1] + [x[2] for x in self.circuit_wires])
            return pos + 1

    def get_circuit_wires(self):
        """Gets a list of all wires in circuit along with their index position for cgp chromosome generation and stores them inside `self.circuit_wires` list.

        Constant wire with value 0 has constant index of 0.
        Constant wire with value 1 has constant index of 1.
        Other wires indexes start counting from 2 and up.
        """
        circuit_wires_names = []

        for input in self.inputs:
            if isinstance(input, Bus):
                [self.circuit_wires.append(
                    (w, f"{w.name}", self.save_wire_id(wire=w))) for w in input.bus]
                [circuit_wires_names.append(w.name) for w in input.bus]
            else:
                self.circuit_wires.append(
                    (input, f"{input.name}", self.save_wire_id(wire=input)))
                circuit_wires_names.append(input.name)

        for gate in self.circuit_gates:
            if gate.a.name not in circuit_wires_names:
                self.circuit_wires.append(
                    (gate.a, gate.a.name, self.save_wire_id(wire=gate.a)))
                circuit_wires_names.append(gate.a.name)

            if hasattr(gate, 'b') and gate.b.name not in circuit_wires_names:
                self.circuit_wires.append(
                    (gate.b, gate.b.name, self.save_wire_id(wire=gate.b)))
                circuit_wires_names.append(gate.b.name)

            if gate.out.name not in circuit_wires_names:
                self.circuit_wires.append(
                    (gate.out, gate.out.name, self.save_wire_id(wire=gate.out)))
                circuit_wires_names.append(gate.out.name)
                

    def get_circuit_wire_index(self, wire: Wire):
        """Searches for circuit's wire unique index position within the circuit. Used for cgp chromosome generation.

        Args:
            wire (Wire): Wire to retrieve index position of.

        Returns:
            int: Wire's index position number within the circuit.
        """
        if wire.is_const():
            return wire.cgp_const
        else:
            for w in self.circuit_wires:
                if wire.name == w[1]:
                    return w[2]

    """ PYTHON CODE GENERATION """
    # FLAT PYTHON #
    def get_prototype_python(self):
        """Generates Python code function header to describe corresponding arithmetic circuit's interface in Python code.

        Returns:
            str: Function's name and parameters in Python code.
        """
        return f"def {self.prefix}(" + ", ".join([f"{x.prefix}" for x in self.inputs]) + ")" + ":" + "\n"

    def get_init_python_flat(self):
        """Generates flat Python code initialization and assignment of corresponding arithmetic circuit's input/output wires.

        Returns:
            str: Flat Python code initialization of arithmetic circuit wires.
        """
        return "".join([c.get_assign_python_flat() if isinstance(c, TwoInputLogicGate) else c.get_init_python_flat() for c in self.components])

    def get_function_out_python_flat(self):
        """Generates flat Python code assignment of corresponding arithmetic circuit's output bus wires.

        Returns:
            str: Flat Python code containing output bus wires assignment.
        """
        return self.out.return_bus_wires_values_python_flat()

    # Generating flat Python code representation of circuit

    def get_python_code_flat(self, file_object, retype=True):
        """Generates flat Python code representation of corresponding arithmetic circuit.

        Args:
            file_object (TextIOWrapper): Destination file object where circuit's representation will be written to.
            retype (bool) specifies if signed output should return int64_t
        """
        file_object.write(self.get_prototype_python())
        file_object.write(self.get_init_python_flat()+"\n")
        file_object.write(self.get_function_out_python_flat())
        file_object.write(self.out.return_bus_wires_sign_extend_python_flat(retype=True))
        file_object.write(f"  return {self.out.prefix}"+"\n")

    """ C CODE GENERATION """
    # FLAT C #
    @staticmethod
    def get_includes_c():
        """Generates necessary C library includes for output representation.

        Returns:
            str: C code library includes.
        """
        return f"#include <stdio.h>\n#include <stdint.h>\n\n"

    def get_prototype_c(self):
        """Generates C code function header to describe corresponding arithmetic circuit's interface in C code.

        Returns:
            str: Function's name and parameters in C code.
        """
        return f"{self.c_data_type} {self.prefix}(" + ",".join([f"{self.c_data_type} {x.prefix}" for x in self.inputs]) + ")" + "{" + "\n"

    def get_declaration_c_flat(self):
        """Generates flat C code declaration of input/output circuit wires.

        Returns:
            str: Flat C code arithmetic circuit's wires declaration.
        """
        return f"".join([c.get_declaration_c_flat() for c in self.components])

    def get_init_c_flat(self):
        """Generates flat C code initialization and assignment of corresponding arithmetic circuit's input/output wires.

        Returns:
            str: Flat C code initialization of arithmetic circuit wires.
        """
        return "".join([c.get_assign_c_flat() if isinstance(c, TwoInputLogicGate) else c.get_init_c_flat() for c in self.components])

    def get_function_out_c_flat(self):
        """Generates flat C code assignment of corresponding arithmetic circuit's output bus wires.

        Returns:
            str: Flat C code containing output bus wires assignment.
        """
        return self.out.return_bus_wires_values_c_flat()

    # Generating flat C code representation of circuit
    def get_c_code_flat(self, file_object):
        """Generates flat C code representation of corresponding arithmetic circuit.

        Args:
            file_object (TextIOWrapper): Destination file object where circuit's representation will be written to.
        """
        file_object.write(self.get_includes_c())
        file_object.write(self.get_prototype_c())
        file_object.write(self.out.get_declaration_c())
        file_object.write(self.get_declaration_c_flat()+"\n")
        file_object.write(self.get_init_c_flat()+"\n")
        file_object.write(self.get_function_out_c_flat())
        file_object.write(self.out.return_bus_wires_sign_extend_c_flat())
        file_object.write(f"  return {self.out.prefix}"+";\n}")

    # HIERARCHICAL C #
    def get_function_blocks_c(self):
        """Generates hierarchical C code representation of all subcomponents function blocks present in corresponding arithmetic circuit.

        Returns:
            str: Hierarchical C code of all subcomponents function blocks description.
        """
        # Retrieve all unique component types composing this circuit and add them kwargs from the parent circuit to allow propagatation of config settings for subcomponents
        self.component_types = self.get_component_types()
        return "".join([c.get_function_block_c() for c in self.component_types])

    def get_function_block_c(self):
        """Generates hierarchical C code representation of corresponding multi-bit arithmetic circuit used as function block in hierarchical circuit description.

        Returns:
            str: Hierarchical C code of multi-bit arithmetic circuit's function block description.
        """
        return f"{self.get_hier_subcomponent_def(parent_kwargs=self.kwargs).get_circuit_c()}\n\n"

    def get_declarations_c_hier(self):
        """Generates hierarchical C code declaration of input/output circuit wires.

        Returns:
            str: Hierarchical C code containing unique declaration of arithmetic circuit wires.
        """
        return "".join([c.get_declaration_c_hier() for c in self.components])

    def get_declaration_c_hier(self):
        """Generates hierarchical C code declaration of corresponding subcomponent input/output wires inside the upper component.

        Generates wires used to connect input/output values to/from invocation of the corresponding function block into inner wires present
        inside the upper component from which function block has been invoked.

        Returns:
            str: Hierarchical C code of subcomponent arithmetic circuit's wires declaration.
        """
        return "".join([f"  {self.c_data_type} {i.prefix} = 0;\n" for i in self.inputs if ((isinstance(i, Wire)) or (not all((w.is_const()) or (w.parent_bus is not None and w.prefix == i.prefix) for w in i.bus)))]) + \
               f"  {self.c_data_type} {self.out.prefix} = 0;\n"

    def get_init_c_hier(self):
        """Generates hierarchical C code initialization and assignment of corresponding arithmetic circuit's input/output wires.

        Returns:
            str: Hierarchical C code initialization of arithmetic circuit wires.
        """
        return "".join([c.get_gate_invocation_c() if isinstance(c, TwoInputLogicGate) else c.get_out_invocation_c() for c in self.components])

    def get_out_invocation_c(self):
        """Generates hierarchical C code invocation of corresponding arithmetic circuit's generated function block.

        Assigns input values from other subcomponents into multi-bit input buses used as inputs for function block invocation.
        Assigns output values from invocation of the corresponding function block into inner wires present inside
        the upper component from which function block has been invoked.

        Returns:
            str: Hierarchical C code of subcomponent's C function invocation and output assignment.
        """
        # Getting name of circuit type for proper C code generation without affecting actual generated composition
        init_signature = inspect.signature(self.__class__.__init__)
        default_circuit_name = init_signature.parameters['name'].default
        circuit_type = default_circuit_name + "x".join(str(getattr(self, chr(97+i)).N) for i, _ in enumerate(self.inputs))
        # TODO .. now only works for input buses
        return "".join(w.return_bus_wires_values_c_hier() for w in self.inputs) + \
               f"  {self.out.prefix} = {circuit_type}({', '.join(w.prefix if isinstance(w, Bus) else w.get_wire_value_c_hier() for w in self.inputs)});\n"

    def get_function_out_c_hier(self):
        """Generates hierarchical C code assignment of corresponding arithmetic circuit's output bus wires.

        Returns:
            str: Hierarchical C code containing output bus wires assignment.
        """
        return self.out.return_bus_wires_values_c_hier()

    def get_circuit_c(self):
        """Generates hierarchical C code subcomponent's function block.

        Returns:
            str: Hierarchical C code of subcomponent's function block.
        """
        return f"{self.get_prototype_c()}" + \
               f"{self.out.get_declaration_c()}" + \
               f"{self.get_declarations_c_hier()}\n" + \
               f"{self.get_init_c_hier()}\n" + \
               f"{self.get_function_out_c_hier()}" + \
               f"{self.out.return_bus_wires_sign_extend_c_hier()}" + \
               f"  return {self.out.prefix}"+";\n}"

    # Generating hierarchical C code representation of circuit
    def get_c_code_hier(self, file_object):
        """Generates hierarchical C code representation of corresponding arithmetic circuit.

        Args:
            file_object (TextIOWrapper): Destination file object where circuit's representation will be written to.
        """
        file_object.write(self.get_includes_c())
        file_object.write(self.get_function_blocks_c())
        file_object.write(self.get_circuit_c())

    """ VERILOG CODE GENERATION """
    # FLAT VERILOG #
    def get_prototype_v(self):
        """Generates Verilog code module header to describe corresponding arithmetic circuit's interface in Verilog code.

        Returns:
            str: Module's name and parameters in Verilog code.
        """
        return f"module {self.prefix}(" + ", ".join(f"input [{x.N-1}:0] {x.prefix}" for x in self.inputs) + f", output [{self.out.N-1}:0] {self.out.prefix});\n"

    def get_declaration_v_flat(self):
        """Generates flat Verilog code declaration of input/output circuit wires.

        Returns:
            str: Flat Verilog code arithmetic circuit's wires declaration.
        """
        return f"".join([c.get_declaration_v_flat() for c in self.components])

    def get_init_v_flat(self):
        """Generates flat Verilog code initialization and assignment of corresponding arithmetic circuit's input/output buses wires.

        Returns:
            str: Flat Verilog code initialization of arithmetic circuit wires.
        """
        return "".join([c.get_assign_v_flat() if isinstance(c, TwoInputLogicGate) else c.get_init_v_flat() for c in self.components])

    def get_function_out_v_flat(self):
        """Generates flat Verilog code assignment of corresponding arithmetic circuit's output bus wires.

        Returns:
            str: Flat Verilog code containing output bus wires assignment.
        """
        return self.out.return_bus_wires_values_v_flat()

    # Generating flat Verilog code representation of circuit
    def get_v_code_flat(self, file_object):
        """Generates flat Verilog code representation of corresponding arithmetic circuit.

        Args:
            file_object (TextIOWrapper): Destination file object where circuit's representation will be written to.
        """
        file_object.write(self.get_prototype_v())
        file_object.write(self.get_declaration_v_flat()+"\n")
        file_object.write(self.get_init_v_flat() + "\n")
        file_object.write(self.get_function_out_v_flat())
        file_object.write(f"endmodule")

    # HIERARCHICAL VERILOG #
    def get_function_blocks_v(self):
        """Generates hierarchical Verilog code representation of all subcomponents function blocks present in corresponding arithmetic circuit.

        Returns:
            str: Hierarchical Verilog code of all subcomponents function blocks description.
        """
        # Retrieve all unique component types composing this circuit and add them kwargs from the parent circuit to allow propagatation of config settings for subcomponents
        self.component_types = self.get_component_types(verilog_output=True)
        return "".join([c.get_function_block_v() for c in self.component_types])

    def get_function_block_v(self):
        """Generates hierarchical Verilog code representation of corresponding multi-bit arithmetic circuit used as function block in hierarchical circuit description.

        Returns:
            str: Hierarchical Verilog code of multi-bit arithmetic circuit's function block description.
        """
        # Obtain proper circuit name with its bit width
        return f"{self.get_hier_subcomponent_def(parent_kwargs=self.kwargs).get_circuit_v()}\n\n"

    def get_declarations_v_hier(self):
        """Generates hierarchical Verilog code declaration of input/output circuit wires.

        Returns:
            str: Hierarchical Verilog code containing unique declaration of arithmetic circuit wires.
        """
        return "".join([c.get_declaration_v_hier() for c in self.components])

    def get_declaration_v_hier(self):
        """Generates hierarchical Verilog code declaration of corresponding subcomponent input/output wires inside the upper component.

        Generates wires used to connect input/output values to/from invocation of the corresponding function block into inner wires present
        inside the upper component from which function block has been invoked.

        Returns:
            str: Hierarchical Verilog code of subcomponent arithmetic circuit's wires declaration.
        """
        return "".join(b.get_wire_declaration_v() for b in self.inputs + [self.out] if (b == self.out) or (not all((w.is_const()) or (w.parent_bus is not None and w.prefix == b.prefix) for w in b.bus)))

    def get_init_v_hier(self):
        """Generates hierarchical Verilog code initialization and assignment of corresponding arithmetic circuit's input/output wires.

        Returns:
            str: Hierarchical Verilog code initialization of arithmetic circuit wires.
        """
        return "".join([c.get_gate_invocation_v() if isinstance(c, TwoInputLogicGate) else c.get_out_invocation_v() for c in self.components])

    def get_out_invocation_v(self):
        """Generates hierarchical Verilog code invocation of corresponding arithmetic circuit's generated function block.

        Assigns input values from other subcomponents into multi-bit input buses used as inputs for function block invocation.
        Assigns output values from invocation of the corresponding function block into inner wires present inside
        the upper component from which function block has been invoked.

        Returns:
            str: Hierarchical Verilog code of subcomponent's module invocation and output assignment.
        """
        # Getting name of circuit type and insitu copying out bus for proper Verilog code generation without affecting actual generated composition
        init_signature = inspect.signature(self.__class__.__init__)
        default_circuit_name = init_signature.parameters['name'].default
        circuit_type = default_circuit_name + "x".join(str(getattr(self, chr(97+i)).N) for i, _ in enumerate(self.inputs))
        circuit_block = self.get_hier_subcomponent_def(parent_kwargs=self.kwargs)
        # TODO .. now only works for input buses
        return "".join([c.return_bus_wires_values_v_hier() for c in self.inputs]) + \
               f"  {circuit_type} {circuit_type}_{self.out.prefix}(" + ",".join([f".{a.prefix}({b.prefix})" for a, b in zip(circuit_block.inputs, self.inputs)]) + f", .{circuit_block.out.prefix}({self.out.prefix}));\n"

    def get_function_out_v_hier(self):
        """Generates hierarchical Verilog code assignment of corresponding arithmetic circuit's output bus wires.

        Returns:
            str: Hierarchical Verilog code containing output bus wires assignment.
        """
        return self.out.return_bus_wires_values_v_hier()

    def get_circuit_v(self):
        """Generates hierarchical Verilog code subcomponent's function block.

        Returns:
            str: Hierarchical Verilog code of subcomponent's function block.
        """
        return f"{self.get_prototype_v()}" + \
               f"{self.get_declarations_v_hier()}\n" + \
               f"{self.get_init_v_hier()}\n" + \
               f"{self.get_function_out_v_hier()}" + \
               f"endmodule"

    # Generating hierarchical Verilog code representation of circuit
    def get_v_code_hier(self, file_object):
        """Generates hierarchical Verilog code representation of corresponding arithmetic circuit.

        Args:
            file_object (TextIOWrapper): Destination file object where circuit's representation will be written to.
        """
        file_object.write(self.get_function_blocks_v())
        file_object.write(self.get_circuit_v())

    """ BLIF CODE GENERATION """
    # FLAT BLIF #

    def get_prototype_blif(self):
        """Generates Blif code model name of described arithmetic circuit.

        Returns:
            str: Model's name in Blif code.
        """
        return f".model {self.prefix}\n"

    def get_declaration_blif(self):
        """Generates flat Blif code declaration of input/output circuit wires.

        Returns:
            str: Flat Blif code containing declaration of circuit's wires.
        """
        return f".inputs {''.join([w.get_wire_declaration_blif() for w in self.inputs])}\n" + \
               f".outputs{self.out.get_wire_declaration_blif()}\n" + \
               f".names vdd\n1\n" + \
               f".names gnd\n0\n"

    def get_function_blif_flat(self):
        """Generates flat Blif code with invocation of subcomponents logic gates functions via their corresponding truth tables.

        Returns:
            str: Flat Blif code containing invocation of inner subcomponents logic gates Boolean functions.
        """
        return "".join(c.get_function_blif_flat() for c in self.components)

    def get_function_out_blif(self):
        """Generates flat Blif code assignment of corresponding arithmetic circuit's output bus wires.

        Returns:
            str: Flat Blif code containing output bus wires assignment.
        """
        return f"{self.out.get_wire_assign_blif(output=True)}"

    # Generating flat BLIF code representation of circuit
    def get_blif_code_flat(self, file_object):
        """Generates flat Blif code representation of corresponding arithmetic circuit.

        Args:
            file_object (TextIOWrapper): Destination file object where circuit's representation will be written to.
        """
        file_object.write(self.get_prototype_blif())
        file_object.write(self.get_declaration_blif())
        file_object.write(self.get_function_blif_flat())
        file_object.write(self.get_function_out_blif())
        file_object.write(f".end\n")









    # HIERARCHICAL BLIF #
    def get_invocations_blif_hier(self):
        """Generates hierarchical Blif code with invocations of subcomponents function blocks.

        Returns:
            str: Hierarchical Blif code containing invocations of inner subcomponents function blocks.
        """
        return "".join(c.get_invocation_blif_hier() for c in self.components)

    def get_invocation_blif_hier(self):
        """Generates hierarchical Blif code invocation of corresponding arithmetic circuit's generated function block.

        Used for multi-bit subcomponent's modul invocation.

        Returns:
            str: Hierarchical Blif code of subcomponent's model invocation and output assignment.
        """
        # Getting name of circuit type for proper Blif code generation without affecting actual generated composition
        init_signature = inspect.signature(self.__class__.__init__)
        default_circuit_name = init_signature.parameters['name'].default
        circuit_type = default_circuit_name + "x".join(str(getattr(self, chr(97+i)).N) for i, _ in enumerate(self.inputs))
        if self.out.N > 1:
            return "".join([w.get_wire_assign_blif(output=True) for w in self.inputs]) + \
                   f".subckt {circuit_type}" + \
                   "".join([f" {chr(97+i)}[{b.bus.index(w)}]={b.prefix}[{b.bus.index(w)}]" if b.N > 1 else f" {chr(97+i)}={b.prefix}" for i, b in enumerate(self.inputs) for w in b.bus]) + \
                   "".join([f" {circuit_type}_out[{self.out.bus.index(o)}]={o.name}" for o in self.out.bus if not o.is_const()]) + "\n"
        else:
            return "".join([w.get_wire_assign_blif(output=True) for w in self.inputs]) + \
                   f".subckt {circuit_type}" + \
                   "".join([f" {chr(97+i)}[{b.bus.index(w)}]={b.prefix}[{b.bus.index(w)}]" if b.N > 1 else f" {chr(97+i)}={b.prefix}" for i, b in enumerate(self.inputs) for w in b.bus]) + \
                   "".join([f" {circuit_type}_out={o.name}" for o in self.out.bus if not o.is_const()]) + "\n"
            
        # TODO delete
        return f"{self.a.get_wire_assign_blif(output=True)}" + \
               f"{self.b.get_wire_assign_blif(output=True)}" + \
               f".subckt {circuit_type}" + \
               "".join([f" a[{self.a.bus.index(w)}]={self.a.prefix}[{self.a.bus.index(w)}]" for w in self.a.bus]) + \
               "".join([f" b[{self.b.bus.index(w)}]={self.b.prefix}[{self.b.bus.index(w)}]" for w in self.b.bus]) + \
               "".join([f" {circuit_type}_out[{self.out.bus.index(o)}]={o.name}" for o in self.out.bus]) + "\n"

    def get_circuit_blif(self):
        """Generates hierarchical Blif code subcomponent's function block.

        Returns:
            str: Hierarchical Blif code of subcomponent's function block.
        """
        return f"{self.get_prototype_blif()}" + \
               f"{self.get_declaration_blif()}" + \
               f"{self.get_invocations_blif_hier()}" + \
               f"{self.get_function_out_blif()}" + \
               f".end\n"

    def get_function_blocks_blif(self):
        """Generates hierarchical Blif code representation of all subcomponents function blocks present in corresponding arithmetic circuit.

        Returns:
            str: Hierarchical Blif code of all subcomponents function blocks description.
        """
        # Retrieve all unique component types composing this circuit and add them kwargs from the parent circuit to allow propagatation of config settings for subcomponents
        # (iterating backwards as opposed to other representations so the top modul is always above its subcomponents)
        self.component_types = self.get_component_types()
        return "\n".join([c.get_function_block_blif() for c in self.component_types[::-1]])

    def get_function_block_blif(self):
        """Generates hierarchical Blif code representation of corresponding multi-bit arithmetic circuit used as function block in hierarchical circuit description.

        Returns:
            str: Hierarchical Blif code of multi-bit arithmetic circuit's function block description.
        """
        # Obtain proper circuit name with its bit width
        return f"{self.get_hier_subcomponent_def(parent_kwargs=self.kwargs).get_circuit_blif()}"

    # Generating hierarchical BLIF code representation of circuit
    def get_blif_code_hier(self, file_object):
        """Generates hierarchical Blif code representation of corresponding arithmetic circuit.

        Args:
            file_object (TextIOWrapper): Destination file object where circuit's representation will be written to.
        """
        file_object.write(self.get_circuit_blif()+"\n")
        file_object.write(self.get_function_blocks_blif())

    """ CGP CODE GENERATION """
    # FLAT CGP #

    def get_parameters_cgp(self):
        """Generates CGP chromosome parameters of corresponding arithmetic circuit.

        In total seven parameters represent: total inputs, total outputs, number of rows, number of columns (gates),
        number of each gate's inputs, number of each gate's outputs, quality constant value.

        Returns:
            str: CGP chromosome parameters of described arithmetic circuit.
        """
        # self.circuit_gates = self.get_circuit_gates() TODO delete
        return f"{{{sum(input_bus.N for input_bus in self.inputs)},{self.out.N},1,{len(self.circuit_gates)},2,1,0}}"

    def get_triplets_cgp(self):
        """Generates list of logic gate triplets (first input wire, second input wire, logic gate function) using wires unique position indexes within the described circuit.

        Each triplet represents unique logic gate within the described arithmetic circuit. Besides the contained input wires indexes and gate's inner logic function, an output wire
        with incremented index position is also created and remembered to be appropriately driven as an input to another logic gate or as the circuit's output.

        Constant wire with value 0 has constant index of 0.
        Constant wire with value 1 has constant index of 1.
        Other wires indexes start counting from 2 and up.

        Returns:
            str: List of triplets each describing logic function of corresponding two input logic gate and as a whole describe the arithmetic circuit.
        """
        self.get_circuit_wires()
        return "".join([g.get_triplet_cgp(a_id=self.get_circuit_wire_index(g.a), out_id=self.get_circuit_wire_index(g.out)) if isinstance(g, OneInputLogicGate) else
                       g.get_triplet_cgp(a_id=self.get_circuit_wire_index(g.a), b_id=self.get_circuit_wire_index(g.b), out_id=self.get_circuit_wire_index(g.out)) for g in self.circuit_gates])

    def get_outputs_cgp(self):
        """Generates list of output wires indexes of described arithmetic circuit from MSB to LSB.

        Returns:
            str: List of arithmetic circuit's output wire indexes.
        """
        return "(" + ",".join([str(self.get_circuit_wire_index(o)) for o in self.out.bus]) + ")"

    # Generating flat CGP chromosome representation of circuit
    def get_cgp_code_flat(self, file_object):
        """Generates flat CGP chromosome representation of corresponding arithmetic circuit.

        Args:
            file_object (TextIOWrapper): Destination file object where circuit's representation will be written to.
        """
        file_object.write(self.get_parameters_cgp())
        file_object.write(self.get_triplets_cgp())
        file_object.write(self.get_outputs_cgp())




import os
import sys
# Add the parent directory to the system path
DIR_PATH = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, os.path.join(DIR_PATH, '..'))
import numpy as np
import math

from ariths_gen.wire_components import (
    Wire,
    ConstantWireValue0,
    ConstantWireValue1,
    Bus
)

from ariths_gen.multi_bit_circuits.multipliers import (
    SignedArrayMultiplier,
    SignedDaddaMultiplier,
    SignedWallaceMultiplier,
    SignedCarrySaveMultiplier
)

from ariths_gen.multi_bit_circuits.adders import (
    UnsignedCarryLookaheadAdder,
    UnsignedPGRippleCarryAdder,
    UnsignedRippleCarryAdder,
    SignedCarryLookaheadAdder,
    SignedPGRippleCarryAdder,
    SignedRippleCarryAdder,
    UnsignedCarrySkipAdder,
    SignedCarrySkipAdder,
    UnsignedKoggeStoneAdder,
    SignedKoggeStoneAdder,
    UnsignedBrentKungAdder,
    SignedBrentKungAdder,
    UnsignedSklanskyAdder,
    SignedSklanskyAdder,
    UnsignedHanCarlsonAdder,
    SignedHanCarlsonAdder,
    UnsignedLadnerFischerAdder,
    SignedLadnerFischerAdder,
    UnsignedKnowlesAdder,
    SignedKnowlesAdder,
    UnsignedCarrySelectAdder,
    SignedCarrySelectAdder,
    UnsignedConditionalSumAdder,
    SignedConditionalSumAdder,
    UnsignedCarryIncrementAdder,
    SignedCarryIncrementAdder
)

def test_unbalanced_signed_multiplier():
    # Example usage

    for mult in [
        SignedDaddaMultiplier,
        SignedArrayMultiplier,
        SignedWallaceMultiplier,
        SignedCarrySaveMultiplier
    ]:
        a = Bus(N=5, prefix="a")
        b = Bus(N=4, prefix="b")
        multiplier = mult(a, b, prefix="sam")

        a_v = np.arange(-15, 16).reshape((-1, 1))
        b_v = np.arange(-7, 8).reshape((1, -1))
        # Test the multiplier circuit
        res = multiplier(a_v, b_v)
        exp = a_v * b_v
        print(res)
        print(res - exp)
        np.testing.assert_allclose(res, exp, rtol=1e-5, atol=1e-5, err_msg=f"Multiplier {mult.__name__} failed for unbalanced inputs a={a_v.shape}, b={b_v.shape}")
    

def test_unbalanced_signed_multiplier2():
    # Example usage

    for mult in [
        SignedDaddaMultiplier,
        SignedArrayMultiplier,
        SignedWallaceMultiplier,
        SignedCarrySaveMultiplier
    ]:
        a = Bus(N=4, prefix="a")
        b = Bus(N=5, prefix="b")
        multiplier = mult(a, b, prefix="sam")

        a_v = np.arange(-7, 8).reshape((-1, 1))
        b_v = np.arange(-15, 16).reshape((1, -1))
        # Test the multiplier circuit
        res = multiplier(a_v, b_v)
        exp = a_v * b_v
        print(res)
        print(res - exp)
        np.testing.assert_allclose(res, exp, rtol=1e-5, atol=1e-5, err_msg=f"Multiplier {mult.__name__} failed for unbalanced inputs a={a_v.shape}, b={b_v.shape}")


def test_unbalanced_signed_adder():
    # Example usage

    for adder in [
        SignedCarryLookaheadAdder,
        SignedPGRippleCarryAdder,
        SignedRippleCarryAdder,
        SignedCarrySkipAdder,
        SignedKoggeStoneAdder,
        SignedBrentKungAdder,
        SignedSklanskyAdder,
        SignedHanCarlsonAdder,
        SignedLadnerFischerAdder,
        SignedKnowlesAdder,
        SignedCarrySelectAdder,
        SignedConditionalSumAdder,
        SignedCarryIncrementAdder
    ]:
        a = Bus(N=5, prefix="a")
        b = Bus(N=4, prefix="b")
        adder_circuit = adder(a, b, prefix="scl")

        a_v = np.arange(-15, 16).reshape((-1, 1))
        b_v = np.arange(-7, 8).reshape((1, -1))
        # Test the adder circuit
        res = adder_circuit(a_v, b_v)
        exp = a_v + b_v
        print(res)
        print(res - exp)
        np.testing.assert_allclose(res, exp, rtol=1e-5, atol=1e-5, err_msg=f"Adder {adder.__name__} failed for unbalanced inputs a={a_v.shape}, b={b_v.shape}")
    
def test_unbalanced_signed_adder2():
    # Example usage

    for adder in [
        SignedCarryLookaheadAdder,
        SignedPGRippleCarryAdder,
        SignedRippleCarryAdder,
        SignedCarrySkipAdder,
        SignedKoggeStoneAdder,
        SignedBrentKungAdder,
        SignedSklanskyAdder,
        SignedHanCarlsonAdder,
        SignedLadnerFischerAdder,
        SignedKnowlesAdder,
        SignedCarrySelectAdder,
        SignedConditionalSumAdder,
        SignedCarryIncrementAdder
    ]:
        a = Bus(N=4, prefix="a")
        b = Bus(N=5, prefix="b")
        adder_circuit = adder(a, b, prefix="scl")

        a_v = np.arange(-7, 8).reshape((-1, 1))
        b_v = np.arange(-15, 16).reshape((1, -1))
        # Test the adder circuit
        res = adder_circuit(a_v, b_v)
        exp = a_v + b_v
        print(res)
        print(res - exp)
        np.testing.assert_allclose(res, exp, rtol=1e-5, atol=1e-5, err_msg=f"Adder {adder.__name__} failed for unbalanced inputs a={a_v.shape}, b={b_v.shape}")